from .helper import PillowTestCase, cached_property

from PIL import Image


class TestImagingPaste(PillowTestCase):
    masks = {}
    size = 128

    def assert_9points_image(self, im, expected):
        expected = [
            point[0] if im.mode == "L" else point[: len(im.mode)] for point in expected
        ]
        px = im.load()
        actual = [
            px[0, 0],
            px[self.size // 2, 0],
            px[self.size - 1, 0],
            px[0, self.size // 2],
            px[self.size // 2, self.size // 2],
            px[self.size - 1, self.size // 2],
            px[0, self.size - 1],
            px[self.size // 2, self.size - 1],
            px[self.size - 1, self.size - 1],
        ]
        self.assertEqual(actual, expected)

    def assert_9points_paste(self, im, im2, mask, expected):
        im3 = im.copy()
        im3.paste(im2, (0, 0), mask)
        self.assert_9points_image(im3, expected)

        # Abbreviated syntax
        im.paste(im2, mask)
        self.assert_9points_image(im, expected)

    @cached_property
    def mask_1(self):
        mask = Image.new("1", (self.size, self.size))
        px = mask.load()
        for y in range(mask.height):
            for x in range(mask.width):
                px[y, x] = (x + y) % 2
        return mask

    @cached_property
    def mask_L(self):
        return self.gradient_L.transpose(Image.ROTATE_270)

    @cached_property
    def gradient_L(self):
        gradient = Image.new("L", (self.size, self.size))
        px = gradient.load()
        for y in range(gradient.height):
            for x in range(gradient.width):
                px[y, x] = (x + y) % 255
        return gradient

    @cached_property
    def gradient_RGB(self):
        return Image.merge(
            "RGB",
            [
                self.gradient_L,
                self.gradient_L.transpose(Image.ROTATE_90),
                self.gradient_L.transpose(Image.ROTATE_180),
            ],
        )

    @cached_property
    def gradient_RGBA(self):
        return Image.merge(
            "RGBA",
            [
                self.gradient_L,
                self.gradient_L.transpose(Image.ROTATE_90),
                self.gradient_L.transpose(Image.ROTATE_180),
                self.gradient_L.transpose(Image.ROTATE_270),
            ],
        )

    @cached_property
    def gradient_RGBa(self):
        return Image.merge(
            "RGBa",
            [
                self.gradient_L,
                self.gradient_L.transpose(Image.ROTATE_90),
                self.gradient_L.transpose(Image.ROTATE_180),
                self.gradient_L.transpose(Image.ROTATE_270),
            ],
        )

    def test_image_solid(self):
        for mode in ("RGBA", "RGB", "L"):
            im = Image.new(mode, (200, 200), "red")
            im2 = getattr(self, "gradient_" + mode)

            im.paste(im2, (12, 23))

            im = im.crop((12, 23, im2.width + 12, im2.height + 23))
            self.assert_image_equal(im, im2)

    def test_image_mask_1(self):
        for mode in ("RGBA", "RGB", "L"):
            im = Image.new(mode, (200, 200), "white")
            im2 = getattr(self, "gradient_" + mode)

            self.assert_9points_paste(
                im,
                im2,
                self.mask_1,
                [
                    (255, 255, 255, 255),
                    (255, 255, 255, 255),
                    (127, 254, 127, 0),
                    (255, 255, 255, 255),
                    (255, 255, 255, 255),
                    (191, 190, 63, 64),
                    (127, 0, 127, 254),
                    (191, 64, 63, 190),
                    (255, 255, 255, 255),
                ],
            )

    def test_image_mask_L(self):
        for mode in ("RGBA", "RGB", "L"):
            im = Image.new(mode, (200, 200), "white")
            im2 = getattr(self, "gradient_" + mode)

            self.assert_9points_paste(
                im,
                im2,
                self.mask_L,
                [
                    (128, 191, 255, 191),
                    (208, 239, 239, 208),
                    (255, 255, 255, 255),
                    (112, 111, 206, 207),
                    (192, 191, 191, 191),
                    (239, 239, 207, 207),
                    (128, 1, 128, 254),
                    (207, 113, 112, 207),
                    (255, 191, 128, 191),
                ],
            )

    def test_image_mask_RGBA(self):
        for mode in ("RGBA", "RGB", "L"):
            im = Image.new(mode, (200, 200), "white")
            im2 = getattr(self, "gradient_" + mode)

            self.assert_9points_paste(
                im,
                im2,
                self.gradient_RGBA,
                [
                    (128, 191, 255, 191),
                    (208, 239, 239, 208),
                    (255, 255, 255, 255),
                    (112, 111, 206, 207),
                    (192, 191, 191, 191),
                    (239, 239, 207, 207),
                    (128, 1, 128, 254),
                    (207, 113, 112, 207),
                    (255, 191, 128, 191),
                ],
            )

    def test_image_mask_RGBa(self):
        for mode in ("RGBA", "RGB", "L"):
            im = Image.new(mode, (200, 200), "white")
            im2 = getattr(self, "gradient_" + mode)

            self.assert_9points_paste(
                im,
                im2,
                self.gradient_RGBa,
                [
                    (128, 255, 126, 255),
                    (0, 127, 126, 255),
                    (126, 253, 126, 255),
                    (128, 127, 254, 255),
                    (0, 255, 254, 255),
                    (126, 125, 254, 255),
                    (128, 1, 128, 255),
                    (0, 129, 128, 255),
                    (126, 255, 128, 255),
                ],
            )

    def test_color_solid(self):
        for mode in ("RGBA", "RGB", "L"):
            im = Image.new(mode, (200, 200), "black")

            rect = (12, 23, 128 + 12, 128 + 23)
            im.paste("white", rect)

            hist = im.crop(rect).histogram()
            while hist:
                head, hist = hist[:256], hist[256:]
                self.assertEqual(head[255], 128 * 128)
                self.assertEqual(sum(head[:255]), 0)

    def test_color_mask_1(self):
        for mode in ("RGBA", "RGB", "L"):
            im = Image.new(mode, (200, 200), (50, 60, 70, 80)[: len(mode)])
            color = (10, 20, 30, 40)[: len(mode)]

            self.assert_9points_paste(
                im,
                color,
                self.mask_1,
                [
                    (50, 60, 70, 80),
                    (50, 60, 70, 80),
                    (10, 20, 30, 40),
                    (50, 60, 70, 80),
                    (50, 60, 70, 80),
                    (10, 20, 30, 40),
                    (10, 20, 30, 40),
                    (10, 20, 30, 40),
                    (50, 60, 70, 80),
                ],
            )

    def test_color_mask_L(self):
        for mode in ("RGBA", "RGB", "L"):
            im = getattr(self, "gradient_" + mode).copy()
            color = "white"

            self.assert_9points_paste(
                im,
                color,
                self.mask_L,
                [
                    (127, 191, 254, 191),
                    (111, 207, 206, 110),
                    (127, 254, 127, 0),
                    (207, 207, 239, 239),
                    (191, 191, 190, 191),
                    (207, 206, 111, 112),
                    (254, 254, 254, 255),
                    (239, 206, 206, 238),
                    (254, 191, 127, 191),
                ],
            )

    def test_color_mask_RGBA(self):
        for mode in ("RGBA", "RGB", "L"):
            im = getattr(self, "gradient_" + mode).copy()
            color = "white"

            self.assert_9points_paste(
                im,
                color,
                self.gradient_RGBA,
                [
                    (127, 191, 254, 191),
                    (111, 207, 206, 110),
                    (127, 254, 127, 0),
                    (207, 207, 239, 239),
                    (191, 191, 190, 191),
                    (207, 206, 111, 112),
                    (254, 254, 254, 255),
                    (239, 206, 206, 238),
                    (254, 191, 127, 191),
                ],
            )

    def test_color_mask_RGBa(self):
        for mode in ("RGBA", "RGB", "L"):
            im = getattr(self, "gradient_" + mode).copy()
            color = "white"

            self.assert_9points_paste(
                im,
                color,
                self.gradient_RGBa,
                [
                    (255, 63, 126, 63),
                    (47, 143, 142, 46),
                    (126, 253, 126, 255),
                    (15, 15, 47, 47),
                    (63, 63, 62, 63),
                    (142, 141, 46, 47),
                    (255, 255, 255, 0),
                    (48, 15, 15, 47),
                    (126, 63, 255, 63),
                ],
            )

    def test_different_sizes(self):
        im = Image.new("RGB", (100, 100))
        im2 = Image.new("RGB", (50, 50))

        im.copy().paste(im2)
        im.copy().paste(im2, (0, 0))



from .helper import PillowTestCase

from PIL import Image
from PIL import ImageFilter

im = Image.open("Tests/images/hopper.ppm")
snakes = Image.open("Tests/images/color_snakes.png")


class TestImageOpsUsm(PillowTestCase):
    def test_filter_api(self):

        test_filter = ImageFilter.GaussianBlur(2.0)
        i = im.filter(test_filter)
        self.assertEqual(i.mode, "RGB")
        self.assertEqual(i.size, (128, 128))

        test_filter = ImageFilter.UnsharpMask(2.0, 125, 8)
        i = im.filter(test_filter)
        self.assertEqual(i.mode, "RGB")
        self.assertEqual(i.size, (128, 128))

    def test_usm_formats(self):

        usm = ImageFilter.UnsharpMask
        self.assertRaises(ValueError, im.convert("1").filter, usm)
        im.convert("L").filter(usm)
        self.assertRaises(ValueError, im.convert("I").filter, usm)
        self.assertRaises(ValueError, im.convert("F").filter, usm)
        im.convert("RGB").filter(usm)
        im.convert("RGBA").filter(usm)
        im.convert("CMYK").filter(usm)
        self.assertRaises(ValueError, im.convert("YCbCr").filter, usm)

    def test_blur_formats(self):

        blur = ImageFilter.GaussianBlur
        self.assertRaises(ValueError, im.convert("1").filter, blur)
        blur(im.convert("L"))
        self.assertRaises(ValueError, im.convert("I").filter, blur)
        self.assertRaises(ValueError, im.convert("F").filter, blur)
        im.convert("RGB").filter(blur)
        im.convert("RGBA").filter(blur)
        im.convert("CMYK").filter(blur)
        self.assertRaises(ValueError, im.convert("YCbCr").filter, blur)

    def test_usm_accuracy(self):

        src = snakes.convert("RGB")
        i = src.filter(ImageFilter.UnsharpMask(5, 1024, 0))
        # Image should not be changed because it have only 0 and 255 levels.
        self.assertEqual(i.tobytes(), src.tobytes())

    def test_blur_accuracy(self):

        i = snakes.filter(ImageFilter.GaussianBlur(0.4))
        # These pixels surrounded with pixels with 255 intensity.
        # They must be very close to 255.
        for x, y, c in [
            (1, 0, 1),
            (2, 0, 1),
            (7, 8, 1),
            (8, 8, 1),
            (2, 9, 1),
            (7, 3, 0),
            (8, 3, 0),
            (5, 8, 0),
            (5, 9, 0),
            (1, 3, 0),
            (4, 3, 2),
            (4, 2, 2),
        ]:
            self.assertGreaterEqual(i.im.getpixel((x, y))[c], 250)
        # Fuzzy match.

        def gp(x, y):
            return i.im.getpixel((x, y))

        self.assertTrue(236 <= gp(7, 4)[0] <= 239)
        self.assertTrue(236 <= gp(7, 5)[2] <= 239)
        self.assertTrue(236 <= gp(7, 6)[2] <= 239)
        self.assertTrue(236 <= gp(7, 7)[1] <= 239)
        self.assertTrue(236 <= gp(8, 4)[0] <= 239)
        self.assertTrue(236 <= gp(8, 5)[2] <= 239)
        self.assertTrue(236 <= gp(8, 6)[2] <= 239)
        self.assertTrue(236 <= gp(8, 7)[1] <= 239)



from .helper import PillowTestCase

from PIL import Image, TarIO

codecs = dir(Image.core)

# Sample tar archive
TEST_TAR_FILE = "Tests/images/hopper.tar"


class TestFileTar(PillowTestCase):
    def setUp(self):
        if "zip_decoder" not in codecs and "jpeg_decoder" not in codecs:
            self.skipTest("neither jpeg nor zip support available")

    def test_sanity(self):
        for codec, test_path, format in [
            ["zip_decoder", "hopper.png", "PNG"],
            ["jpeg_decoder", "hopper.jpg", "JPEG"],
        ]:
            if codec in codecs:
                tar = TarIO.TarIO(TEST_TAR_FILE, test_path)
                im = Image.open(tar)
                im.load()
                self.assertEqual(im.mode, "RGB")
                self.assertEqual(im.size, (128, 128))
                self.assertEqual(im.format, format)

    def test_close(self):
        tar = TarIO.TarIO(TEST_TAR_FILE, "hopper.jpg")
        tar.close()

    def test_contextmanager(self):
        with TarIO.TarIO(TEST_TAR_FILE, "hopper.jpg"):
            pass



#
# The Python Imaging Library.
# $Id$
#
# Windows Icon support for PIL
#
# History:
#       96-05-27 fl     Created
#
# Copyright (c) Secret Labs AB 1997.
# Copyright (c) Fredrik Lundh 1996.
#
# See the README file for information on usage and redistribution.
#

# This plugin is a refactored version of Win32IconImagePlugin by Bryan Davis
# <casadebender@gmail.com>.
# https://code.google.com/archive/p/casadebender/wikis/Win32IconImagePlugin.wiki
#
# Icon format references:
#   * https://en.wikipedia.org/wiki/ICO_(file_format)
#   * https://msdn.microsoft.com/en-us/library/ms997538.aspx


import struct
import warnings
from io import BytesIO

from . import Image, ImageFile, BmpImagePlugin, PngImagePlugin
from ._binary import i8, i16le as i16, i32le as i32
from math import log, ceil

# __version__ is deprecated and will be removed in a future version. Use
# PIL.__version__ instead.
__version__ = "0.1"

#
# --------------------------------------------------------------------

_MAGIC = b"\0\0\1\0"


def _save(im, fp, filename):
    fp.write(_MAGIC)  # (2+2)
    sizes = im.encoderinfo.get(
        "sizes",
        [(16, 16), (24, 24), (32, 32), (48, 48), (64, 64), (128, 128), (256, 256)],
    )
    width, height = im.size
    sizes = filter(
        lambda x: False
        if (x[0] > width or x[1] > height or x[0] > 256 or x[1] > 256)
        else True,
        sizes,
    )
    sizes = list(sizes)
    fp.write(struct.pack("<H", len(sizes)))  # idCount(2)
    offset = fp.tell() + len(sizes) * 16
    for size in sizes:
        width, height = size
        # 0 means 256
        fp.write(struct.pack("B", width if width < 256 else 0))  # bWidth(1)
        fp.write(struct.pack("B", height if height < 256 else 0))  # bHeight(1)
        fp.write(b"\0")  # bColorCount(1)
        fp.write(b"\0")  # bReserved(1)
        fp.write(b"\0\0")  # wPlanes(2)
        fp.write(struct.pack("<H", 32))  # wBitCount(2)

        image_io = BytesIO()
        tmp = im.copy()
        tmp.thumbnail(size, Image.LANCZOS)
        tmp.save(image_io, "png")
        image_io.seek(0)
        image_bytes = image_io.read()
        bytes_len = len(image_bytes)
        fp.write(struct.pack("<I", bytes_len))  # dwBytesInRes(4)
        fp.write(struct.pack("<I", offset))  # dwImageOffset(4)
        current = fp.tell()
        fp.seek(offset)
        fp.write(image_bytes)
        offset = offset + bytes_len
        fp.seek(current)


def _accept(prefix):
    return prefix[:4] == _MAGIC


class IcoFile(object):
    def __init__(self, buf):
        """
        Parse image from file-like object containing ico file data
        """

        # check magic
        s = buf.read(6)
        if not _accept(s):
            raise SyntaxError("not an ICO file")

        self.buf = buf
        self.entry = []

        # Number of items in file
        self.nb_items = i16(s[4:])

        # Get headers for each item
        for i in range(self.nb_items):
            s = buf.read(16)

            icon_header = {
                "width": i8(s[0]),
                "height": i8(s[1]),
                "nb_color": i8(s[2]),  # No. of colors in image (0 if >=8bpp)
                "reserved": i8(s[3]),
                "planes": i16(s[4:]),
                "bpp": i16(s[6:]),
                "size": i32(s[8:]),
                "offset": i32(s[12:]),
            }

            # See Wikipedia
            for j in ("width", "height"):
                if not icon_header[j]:
                    icon_header[j] = 256

            # See Wikipedia notes about color depth.
            # We need this just to differ images with equal sizes
            icon_header["color_depth"] = (
                icon_header["bpp"]
                or (
                    icon_header["nb_color"] != 0
                    and ceil(log(icon_header["nb_color"], 2))
                )
                or 256
            )

            icon_header["dim"] = (icon_header["width"], icon_header["height"])
            icon_header["square"] = icon_header["width"] * icon_header["height"]

            self.entry.append(icon_header)

        self.entry = sorted(self.entry, key=lambda x: x["color_depth"])
        # ICO images are usually squares
        # self.entry = sorted(self.entry, key=lambda x: x['width'])
        self.entry = sorted(self.entry, key=lambda x: x["square"])
        self.entry.reverse()

    def sizes(self):
        """
        Get a list of all available icon sizes and color depths.
        """
        return {(h["width"], h["height"]) for h in self.entry}

    def getentryindex(self, size, bpp=False):
        for (i, h) in enumerate(self.entry):
            if size == h["dim"] and (bpp is False or bpp == h["color_depth"]):
                return i
        return 0

    def getimage(self, size, bpp=False):
        """
        Get an image from the icon
        """
        return self.frame(self.getentryindex(size, bpp))

    def frame(self, idx):
        """
        Get an image from frame idx
        """

        header = self.entry[idx]

        self.buf.seek(header["offset"])
        data = self.buf.read(8)
        self.buf.seek(header["offset"])

        if data[:8] == PngImagePlugin._MAGIC:
            # png frame
            im = PngImagePlugin.PngImageFile(self.buf)
        else:
            # XOR + AND mask bmp frame
            im = BmpImagePlugin.DibImageFile(self.buf)

            # change tile dimension to only encompass XOR image
            im._size = (im.size[0], int(im.size[1] / 2))
            d, e, o, a = im.tile[0]
            im.tile[0] = d, (0, 0) + im.size, o, a

            # figure out where AND mask image starts
            mode = a[0]
            bpp = 8
            for k, v in BmpImagePlugin.BIT2MODE.items():
                if mode == v[1]:
                    bpp = k
                    break

            if 32 == bpp:
                # 32-bit color depth icon image allows semitransparent areas
                # PIL's DIB format ignores transparency bits, recover them.
                # The DIB is packed in BGRX byte order where X is the alpha
                # channel.

                # Back up to start of bmp data
                self.buf.seek(o)
                # extract every 4th byte (eg. 3,7,11,15,...)
                alpha_bytes = self.buf.read(im.size[0] * im.size[1] * 4)[3::4]

                # convert to an 8bpp grayscale image
                mask = Image.frombuffer(
                    "L",  # 8bpp
                    im.size,  # (w, h)
                    alpha_bytes,  # source chars
                    "raw",  # raw decoder
                    ("L", 0, -1),  # 8bpp inverted, unpadded, reversed
                )
            else:
                # get AND image from end of bitmap
                w = im.size[0]
                if (w % 32) > 0:
                    # bitmap row data is aligned to word boundaries
                    w += 32 - (im.size[0] % 32)

                # the total mask data is
                # padded row size * height / bits per char

                and_mask_offset = o + int(im.size[0] * im.size[1] * (bpp / 8.0))
                total_bytes = int((w * im.size[1]) / 8)

                self.buf.seek(and_mask_offset)
                mask_data = self.buf.read(total_bytes)

                # convert raw data to image
                mask = Image.frombuffer(
                    "1",  # 1 bpp
                    im.size,  # (w, h)
                    mask_data,  # source chars
                    "raw",  # raw decoder
                    ("1;I", int(w / 8), -1),  # 1bpp inverted, padded, reversed
                )

                # now we have two images, im is XOR image and mask is AND image

            # apply mask image as alpha channel
            im = im.convert("RGBA")
            im.putalpha(mask)

        return im


##
# Image plugin for Windows Icon files.


class IcoImageFile(ImageFile.ImageFile):
    """
    PIL read-only image support for Microsoft Windows .ico files.

    By default the largest resolution image in the file will be loaded. This
    can be changed by altering the 'size' attribute before calling 'load'.

    The info dictionary has a key 'sizes' that is a list of the sizes available
    in the icon file.

    Handles classic, XP and Vista icon formats.

    This plugin is a refactored version of Win32IconImagePlugin by Bryan Davis
    <casadebender@gmail.com>.
    https://code.google.com/archive/p/casadebender/wikis/Win32IconImagePlugin.wiki
    """

    format = "ICO"
    format_description = "Windows Icon"

    def _open(self):
        self.ico = IcoFile(self.fp)
        self.info["sizes"] = self.ico.sizes()
        self.size = self.ico.entry[0]["dim"]
        self.load()

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value):
        if value not in self.info["sizes"]:
            raise ValueError("This is not one of the allowed sizes of this image")
        self._size = value

    def load(self):
        if self.im and self.im.size == self.size:
            # Already loaded
            return
        im = self.ico.getimage(self.size)
        # if tile is PNG, it won't really be loaded yet
        im.load()
        self.im = im.im
        self.mode = im.mode
        if im.size != self.size:
            warnings.warn("Image was not the expected size")

            index = self.ico.getentryindex(self.size)
            sizes = list(self.info["sizes"])
            sizes[index] = im.size
            self.info["sizes"] = set(sizes)

            self.size = im.size

    def load_seek(self):
        # Flag the ImageFile.Parser so that it
        # just does all the decode at the end.
        pass


#
# --------------------------------------------------------------------


Image.register_open(IcoImageFile.format, IcoImageFile, _accept)
Image.register_save(IcoImageFile.format, _save)
Image.register_extension(IcoImageFile.format, ".ico")

Image.register_mime(IcoImageFile.format, "image/x-icon")



# Tests potential DOS of Jpeg2kImagePlugin with 0 length block.
# Run from anywhere that PIL is importable.

from PIL import Image
from PIL._util import py3
from io import BytesIO

if py3:
    Image.open(
        BytesIO(
            bytes(
                "\x00\x00\x00\x0cjP\x20\x20\x0d\x0a\x87\x0a\x00\x00\x00\x00hang",
                "latin-1",
            )
        )
    )

else:
    Image.open(
        BytesIO(bytes("\x00\x00\x00\x0cjP\x20\x20\x0d\x0a\x87\x0a\x00\x00\x00\x00hang"))
    )



from .helper import PillowTestCase, hopper

from PIL import Image, IptcImagePlugin

TEST_FILE = "Tests/images/iptc.jpg"


class TestFileIptc(PillowTestCase):
    def test_getiptcinfo_jpg_none(self):
        # Arrange
        im = hopper()

        # Act
        iptc = IptcImagePlugin.getiptcinfo(im)

        # Assert
        self.assertIsNone(iptc)

    def test_getiptcinfo_jpg_found(self):
        # Arrange
        im = Image.open(TEST_FILE)

        # Act
        iptc = IptcImagePlugin.getiptcinfo(im)

        # Assert
        self.assertIsInstance(iptc, dict)
        self.assertEqual(iptc[(2, 90)], b"Budapest")
        self.assertEqual(iptc[(2, 101)], b"Hungary")

    def test_getiptcinfo_tiff_none(self):
        # Arrange
        im = Image.open("Tests/images/hopper.tif")

        # Act
        iptc = IptcImagePlugin.getiptcinfo(im)

        # Assert
        self.assertIsNone(iptc)

    def test_i(self):
        # Arrange
        c = b"a"

        # Act
        ret = IptcImagePlugin.i(c)

        # Assert
        self.assertEqual(ret, 97)

    def test_dump(self):
        # Arrange
        c = b"abc"
        # Temporarily redirect stdout
        try:
            from cStringIO import StringIO
        except ImportError:
            from io import StringIO
        import sys

        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()

        # Act
        IptcImagePlugin.dump(c)

        # Reset stdout
        sys.stdout = old_stdout

        # Assert
        self.assertEqual(mystdout.getvalue(), "61 62 63 \n")



#
# Python Imaging Library
# $Id$
#
# stuff to read GIMP palette files
#
# History:
# 1997-08-23 fl     Created
# 2004-09-07 fl     Support GIMP 2.0 palette files.
#
# Copyright (c) Secret Labs AB 1997-2004.  All rights reserved.
# Copyright (c) Fredrik Lundh 1997-2004.
#
# See the README file for information on usage and redistribution.
#

import re
from ._binary import o8


##
# File handler for GIMP's palette format.


class GimpPaletteFile(object):

    rawmode = "RGB"

    def __init__(self, fp):

        self.palette = [o8(i) * 3 for i in range(256)]

        if fp.readline()[:12] != b"GIMP Palette":
            raise SyntaxError("not a GIMP palette file")

        for i in range(256):

            s = fp.readline()
            if not s:
                break

            # skip fields and comment lines
            if re.match(br"\w+:|#", s):
                continue
            if len(s) > 100:
                raise SyntaxError("bad palette file")

            v = tuple(map(int, s.split()[:3]))
            if len(v) != 3:
                raise ValueError("bad palette entry")

            self.palette[i] = o8(v[0]) + o8(v[1]) + o8(v[2])

        self.palette = b"".join(self.palette)

    def getpalette(self):

        return self.palette, self.rawmode



from .helper import PillowTestCase, hopper
from .test_imageqt import PillowQtTestCase

from PIL import ImageQt, Image


class TestFromQImage(PillowQtTestCase, PillowTestCase):

    files_to_test = [
        hopper(),
        Image.open("Tests/images/transparent.png"),
        Image.open("Tests/images/7x13.png"),
    ]

    def roundtrip(self, expected):
        # PIL -> Qt
        intermediate = expected.toqimage()
        # Qt -> PIL
        result = ImageQt.fromqimage(intermediate)

        if intermediate.hasAlphaChannel():
            self.assert_image_equal(result, expected.convert("RGBA"))
        else:
            self.assert_image_equal(result, expected.convert("RGB"))

    def test_sanity_1(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("1"))

    def test_sanity_rgb(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("RGB"))

    def test_sanity_rgba(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("RGBA"))

    def test_sanity_l(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("L"))

    def test_sanity_p(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("P"))



#
# The Python Imaging Library.
# $Id$
#
# TIFF tags
#
# This module provides clear-text names for various well-known
# TIFF tags.  the TIFF codec works just fine without it.
#
# Copyright (c) Secret Labs AB 1999.
#
# See the README file for information on usage and redistribution.
#

##
# This module provides constants and clear-text names for various
# well-known TIFF tags.
##

from collections import namedtuple


class TagInfo(namedtuple("_TagInfo", "value name type length enum")):
    __slots__ = []

    def __new__(cls, value=None, name="unknown", type=None, length=None, enum=None):
        return super(TagInfo, cls).__new__(cls, value, name, type, length, enum or {})

    def cvt_enum(self, value):
        # Using get will call hash(value), which can be expensive
        # for some types (e.g. Fraction). Since self.enum is rarely
        # used, it's usually better to test it first.
        return self.enum.get(value, value) if self.enum else value


def lookup(tag):
    """
    :param tag: Integer tag number
    :returns: Taginfo namedtuple, From the TAGS_V2 info if possible,
        otherwise just populating the value and name from TAGS.
        If the tag is not recognized, "unknown" is returned for the name

    """

    return TAGS_V2.get(tag, TagInfo(tag, TAGS.get(tag, "unknown")))


##
# Map tag numbers to tag info.
#
#  id: (Name, Type, Length, enum_values)
#
# The length here differs from the length in the tiff spec.  For
# numbers, the tiff spec is for the number of fields returned. We
# agree here.  For string-like types, the tiff spec uses the length of
# field in bytes.  In Pillow, we are using the number of expected
# fields, in general 1 for string-like types.


BYTE = 1
ASCII = 2
SHORT = 3
LONG = 4
RATIONAL = 5
SIGNED_BYTE = 6
UNDEFINED = 7
SIGNED_SHORT = 8
SIGNED_LONG = 9
SIGNED_RATIONAL = 10
FLOAT = 11
DOUBLE = 12

TAGS_V2 = {
    254: ("NewSubfileType", LONG, 1),
    255: ("SubfileType", SHORT, 1),
    256: ("ImageWidth", LONG, 1),
    257: ("ImageLength", LONG, 1),
    258: ("BitsPerSample", SHORT, 0),
    259: (
        "Compression",
        SHORT,
        1,
        {
            "Uncompressed": 1,
            "CCITT 1d": 2,
            "Group 3 Fax": 3,
            "Group 4 Fax": 4,
            "LZW": 5,
            "JPEG": 6,
            "PackBits": 32773,
        },
    ),
    262: (
        "PhotometricInterpretation",
        SHORT,
        1,
        {
            "WhiteIsZero": 0,
            "BlackIsZero": 1,
            "RGB": 2,
            "RGB Palette": 3,
            "Transparency Mask": 4,
            "CMYK": 5,
            "YCbCr": 6,
            "CieLAB": 8,
            "CFA": 32803,  # TIFF/EP, Adobe DNG
            "LinearRaw": 32892,  # Adobe DNG
        },
    ),
    263: ("Threshholding", SHORT, 1),
    264: ("CellWidth", SHORT, 1),
    265: ("CellLength", SHORT, 1),
    266: ("FillOrder", SHORT, 1),
    269: ("DocumentName", ASCII, 1),
    270: ("ImageDescription", ASCII, 1),
    271: ("Make", ASCII, 1),
    272: ("Model", ASCII, 1),
    273: ("StripOffsets", LONG, 0),
    274: ("Orientation", SHORT, 1),
    277: ("SamplesPerPixel", SHORT, 1),
    278: ("RowsPerStrip", LONG, 1),
    279: ("StripByteCounts", LONG, 0),
    280: ("MinSampleValue", LONG, 0),
    281: ("MaxSampleValue", SHORT, 0),
    282: ("XResolution", RATIONAL, 1),
    283: ("YResolution", RATIONAL, 1),
    284: ("PlanarConfiguration", SHORT, 1, {"Contiguous": 1, "Separate": 2}),
    285: ("PageName", ASCII, 1),
    286: ("XPosition", RATIONAL, 1),
    287: ("YPosition", RATIONAL, 1),
    288: ("FreeOffsets", LONG, 1),
    289: ("FreeByteCounts", LONG, 1),
    290: ("GrayResponseUnit", SHORT, 1),
    291: ("GrayResponseCurve", SHORT, 0),
    292: ("T4Options", LONG, 1),
    293: ("T6Options", LONG, 1),
    296: ("ResolutionUnit", SHORT, 1, {"none": 1, "inch": 2, "cm": 3}),
    297: ("PageNumber", SHORT, 2),
    301: ("TransferFunction", SHORT, 0),
    305: ("Software", ASCII, 1),
    306: ("DateTime", ASCII, 1),
    315: ("Artist", ASCII, 1),
    316: ("HostComputer", ASCII, 1),
    317: ("Predictor", SHORT, 1, {"none": 1, "Horizontal Differencing": 2}),
    318: ("WhitePoint", RATIONAL, 2),
    319: ("PrimaryChromaticities", RATIONAL, 6),
    320: ("ColorMap", SHORT, 0),
    321: ("HalftoneHints", SHORT, 2),
    322: ("TileWidth", LONG, 1),
    323: ("TileLength", LONG, 1),
    324: ("TileOffsets", LONG, 0),
    325: ("TileByteCounts", LONG, 0),
    332: ("InkSet", SHORT, 1),
    333: ("InkNames", ASCII, 1),
    334: ("NumberOfInks", SHORT, 1),
    336: ("DotRange", SHORT, 0),
    337: ("TargetPrinter", ASCII, 1),
    338: ("ExtraSamples", SHORT, 0),
    339: ("SampleFormat", SHORT, 0),
    340: ("SMinSampleValue", DOUBLE, 0),
    341: ("SMaxSampleValue", DOUBLE, 0),
    342: ("TransferRange", SHORT, 6),
    347: ("JPEGTables", UNDEFINED, 1),
    # obsolete JPEG tags
    512: ("JPEGProc", SHORT, 1),
    513: ("JPEGInterchangeFormat", LONG, 1),
    514: ("JPEGInterchangeFormatLength", LONG, 1),
    515: ("JPEGRestartInterval", SHORT, 1),
    517: ("JPEGLosslessPredictors", SHORT, 0),
    518: ("JPEGPointTransforms", SHORT, 0),
    519: ("JPEGQTables", LONG, 0),
    520: ("JPEGDCTables", LONG, 0),
    521: ("JPEGACTables", LONG, 0),
    529: ("YCbCrCoefficients", RATIONAL, 3),
    530: ("YCbCrSubSampling", SHORT, 2),
    531: ("YCbCrPositioning", SHORT, 1),
    532: ("ReferenceBlackWhite", RATIONAL, 6),
    700: ("XMP", BYTE, 1),
    33432: ("Copyright", ASCII, 1),
    34377: ("PhotoshopInfo", BYTE, 1),
    # FIXME add more tags here
    34665: ("ExifIFD", SHORT, 1),
    34675: ("ICCProfile", UNDEFINED, 1),
    34853: ("GPSInfoIFD", BYTE, 1),
    # MPInfo
    45056: ("MPFVersion", UNDEFINED, 1),
    45057: ("NumberOfImages", LONG, 1),
    45058: ("MPEntry", UNDEFINED, 1),
    45059: ("ImageUIDList", UNDEFINED, 0),  # UNDONE, check
    45060: ("TotalFrames", LONG, 1),
    45313: ("MPIndividualNum", LONG, 1),
    45569: ("PanOrientation", LONG, 1),
    45570: ("PanOverlap_H", RATIONAL, 1),
    45571: ("PanOverlap_V", RATIONAL, 1),
    45572: ("BaseViewpointNum", LONG, 1),
    45573: ("ConvergenceAngle", SIGNED_RATIONAL, 1),
    45574: ("BaselineLength", RATIONAL, 1),
    45575: ("VerticalDivergence", SIGNED_RATIONAL, 1),
    45576: ("AxisDistance_X", SIGNED_RATIONAL, 1),
    45577: ("AxisDistance_Y", SIGNED_RATIONAL, 1),
    45578: ("AxisDistance_Z", SIGNED_RATIONAL, 1),
    45579: ("YawAngle", SIGNED_RATIONAL, 1),
    45580: ("PitchAngle", SIGNED_RATIONAL, 1),
    45581: ("RollAngle", SIGNED_RATIONAL, 1),
    50741: ("MakerNoteSafety", SHORT, 1, {"Unsafe": 0, "Safe": 1}),
    50780: ("BestQualityScale", RATIONAL, 1),
    50838: ("ImageJMetaDataByteCounts", LONG, 0),  # Can be more than one
    50839: ("ImageJMetaData", UNDEFINED, 1),  # see Issue #2006
}

# Legacy Tags structure
# these tags aren't included above, but were in the previous versions
TAGS = {
    347: "JPEGTables",
    700: "XMP",
    # Additional Exif Info
    32932: "Wang Annotation",
    33434: "ExposureTime",
    33437: "FNumber",
    33445: "MD FileTag",
    33446: "MD ScalePixel",
    33447: "MD ColorTable",
    33448: "MD LabName",
    33449: "MD SampleInfo",
    33450: "MD PrepDate",
    33451: "MD PrepTime",
    33452: "MD FileUnits",
    33550: "ModelPixelScaleTag",
    33723: "IptcNaaInfo",
    33918: "INGR Packet Data Tag",
    33919: "INGR Flag Registers",
    33920: "IrasB Transformation Matrix",
    33922: "ModelTiepointTag",
    34264: "ModelTransformationTag",
    34377: "PhotoshopInfo",
    34735: "GeoKeyDirectoryTag",
    34736: "GeoDoubleParamsTag",
    34737: "GeoAsciiParamsTag",
    34850: "ExposureProgram",
    34852: "SpectralSensitivity",
    34855: "ISOSpeedRatings",
    34856: "OECF",
    34864: "SensitivityType",
    34865: "StandardOutputSensitivity",
    34866: "RecommendedExposureIndex",
    34867: "ISOSpeed",
    34868: "ISOSpeedLatitudeyyy",
    34869: "ISOSpeedLatitudezzz",
    34908: "HylaFAX FaxRecvParams",
    34909: "HylaFAX FaxSubAddress",
    34910: "HylaFAX FaxRecvTime",
    36864: "ExifVersion",
    36867: "DateTimeOriginal",
    36868: "DateTImeDigitized",
    37121: "ComponentsConfiguration",
    37122: "CompressedBitsPerPixel",
    37724: "ImageSourceData",
    37377: "ShutterSpeedValue",
    37378: "ApertureValue",
    37379: "BrightnessValue",
    37380: "ExposureBiasValue",
    37381: "MaxApertureValue",
    37382: "SubjectDistance",
    37383: "MeteringMode",
    37384: "LightSource",
    37385: "Flash",
    37386: "FocalLength",
    37396: "SubjectArea",
    37500: "MakerNote",
    37510: "UserComment",
    37520: "SubSec",
    37521: "SubSecTimeOriginal",
    37522: "SubsecTimeDigitized",
    40960: "FlashPixVersion",
    40961: "ColorSpace",
    40962: "PixelXDimension",
    40963: "PixelYDimension",
    40964: "RelatedSoundFile",
    40965: "InteroperabilityIFD",
    41483: "FlashEnergy",
    41484: "SpatialFrequencyResponse",
    41486: "FocalPlaneXResolution",
    41487: "FocalPlaneYResolution",
    41488: "FocalPlaneResolutionUnit",
    41492: "SubjectLocation",
    41493: "ExposureIndex",
    41495: "SensingMethod",
    41728: "FileSource",
    41729: "SceneType",
    41730: "CFAPattern",
    41985: "CustomRendered",
    41986: "ExposureMode",
    41987: "WhiteBalance",
    41988: "DigitalZoomRatio",
    41989: "FocalLengthIn35mmFilm",
    41990: "SceneCaptureType",
    41991: "GainControl",
    41992: "Contrast",
    41993: "Saturation",
    41994: "Sharpness",
    41995: "DeviceSettingDescription",
    41996: "SubjectDistanceRange",
    42016: "ImageUniqueID",
    42032: "CameraOwnerName",
    42033: "BodySerialNumber",
    42034: "LensSpecification",
    42035: "LensMake",
    42036: "LensModel",
    42037: "LensSerialNumber",
    42112: "GDAL_METADATA",
    42113: "GDAL_NODATA",
    42240: "Gamma",
    50215: "Oce Scanjob Description",
    50216: "Oce Application Selector",
    50217: "Oce Identification Number",
    50218: "Oce ImageLogic Characteristics",
    # Adobe DNG
    50706: "DNGVersion",
    50707: "DNGBackwardVersion",
    50708: "UniqueCameraModel",
    50709: "LocalizedCameraModel",
    50710: "CFAPlaneColor",
    50711: "CFALayout",
    50712: "LinearizationTable",
    50713: "BlackLevelRepeatDim",
    50714: "BlackLevel",
    50715: "BlackLevelDeltaH",
    50716: "BlackLevelDeltaV",
    50717: "WhiteLevel",
    50718: "DefaultScale",
    50719: "DefaultCropOrigin",
    50720: "DefaultCropSize",
    50721: "ColorMatrix1",
    50722: "ColorMatrix2",
    50723: "CameraCalibration1",
    50724: "CameraCalibration2",
    50725: "ReductionMatrix1",
    50726: "ReductionMatrix2",
    50727: "AnalogBalance",
    50728: "AsShotNeutral",
    50729: "AsShotWhiteXY",
    50730: "BaselineExposure",
    50731: "BaselineNoise",
    50732: "BaselineSharpness",
    50733: "BayerGreenSplit",
    50734: "LinearResponseLimit",
    50735: "CameraSerialNumber",
    50736: "LensInfo",
    50737: "ChromaBlurRadius",
    50738: "AntiAliasStrength",
    50740: "DNGPrivateData",
    50778: "CalibrationIlluminant1",
    50779: "CalibrationIlluminant2",
    50784: "Alias Layer Metadata",
}


def _populate():
    for k, v in TAGS_V2.items():
        # Populate legacy structure.
        TAGS[k] = v[0]
        if len(v) == 4:
            for sk, sv in v[3].items():
                TAGS[(k, sv)] = sk

        TAGS_V2[k] = TagInfo(k, *v)


_populate()
##
# Map type numbers to type names -- defined in ImageFileDirectory.

TYPES = {}

# was:
# TYPES = {
#     1: "byte",
#     2: "ascii",
#     3: "short",
#     4: "long",
#     5: "rational",
#     6: "signed byte",
#     7: "undefined",
#     8: "signed short",
#     9: "signed long",
#     10: "signed rational",
#     11: "float",
#     12: "double",
# }

#
# These tags are handled by default in libtiff, without
# adding to the custom dictionary. From tif_dir.c, searching for
# case TIFFTAG in the _TIFFVSetField function:
# Line: item.
# 148: case TIFFTAG_SUBFILETYPE:
# 151: case TIFFTAG_IMAGEWIDTH:
# 154: case TIFFTAG_IMAGELENGTH:
# 157: case TIFFTAG_BITSPERSAMPLE:
# 181: case TIFFTAG_COMPRESSION:
# 202: case TIFFTAG_PHOTOMETRIC:
# 205: case TIFFTAG_THRESHHOLDING:
# 208: case TIFFTAG_FILLORDER:
# 214: case TIFFTAG_ORIENTATION:
# 221: case TIFFTAG_SAMPLESPERPIXEL:
# 228: case TIFFTAG_ROWSPERSTRIP:
# 238: case TIFFTAG_MINSAMPLEVALUE:
# 241: case TIFFTAG_MAXSAMPLEVALUE:
# 244: case TIFFTAG_SMINSAMPLEVALUE:
# 247: case TIFFTAG_SMAXSAMPLEVALUE:
# 250: case TIFFTAG_XRESOLUTION:
# 256: case TIFFTAG_YRESOLUTION:
# 262: case TIFFTAG_PLANARCONFIG:
# 268: case TIFFTAG_XPOSITION:
# 271: case TIFFTAG_YPOSITION:
# 274: case TIFFTAG_RESOLUTIONUNIT:
# 280: case TIFFTAG_PAGENUMBER:
# 284: case TIFFTAG_HALFTONEHINTS:
# 288: case TIFFTAG_COLORMAP:
# 294: case TIFFTAG_EXTRASAMPLES:
# 298: case TIFFTAG_MATTEING:
# 305: case TIFFTAG_TILEWIDTH:
# 316: case TIFFTAG_TILELENGTH:
# 327: case TIFFTAG_TILEDEPTH:
# 333: case TIFFTAG_DATATYPE:
# 344: case TIFFTAG_SAMPLEFORMAT:
# 361: case TIFFTAG_IMAGEDEPTH:
# 364: case TIFFTAG_SUBIFD:
# 376: case TIFFTAG_YCBCRPOSITIONING:
# 379: case TIFFTAG_YCBCRSUBSAMPLING:
# 383: case TIFFTAG_TRANSFERFUNCTION:
# 389: case TIFFTAG_REFERENCEBLACKWHITE:
# 393: case TIFFTAG_INKNAMES:

# Following pseudo-tags are also handled by default in libtiff:
# TIFFTAG_JPEGQUALITY 65537

# some of these are not in our TAGS_V2 dict and were included from tiff.h

# This list also exists in encode.c
LIBTIFF_CORE = {
    255,
    256,
    257,
    258,
    259,
    262,
    263,
    266,
    274,
    277,
    278,
    280,
    281,
    340,
    341,
    282,
    283,
    284,
    286,
    287,
    296,
    297,
    321,
    320,
    338,
    32995,
    322,
    323,
    32998,
    32996,
    339,
    32997,
    330,
    531,
    530,
    301,
    532,
    333,
    # as above
    269,  # this has been in our tests forever, and works
    65537,
}

LIBTIFF_CORE.remove(320)  # Array of short, crashes
LIBTIFF_CORE.remove(301)  # Array of short, crashes
LIBTIFF_CORE.remove(532)  # Array of long, crashes

LIBTIFF_CORE.remove(255)  # We don't have support for subfiletypes
LIBTIFF_CORE.remove(322)  # We don't have support for writing tiled images with libtiff
LIBTIFF_CORE.remove(323)  # Tiled images
LIBTIFF_CORE.remove(333)  # Ink Names either

# Note to advanced users: There may be combinations of these
# parameters and values that when added properly, will work and
# produce valid tiff images that may work in your application.
# It is safe to add and remove tags from this set from Pillow's point
# of view so long as you test against libtiff.



from __future__ import print_function
from .helper import PillowTestCase

from PIL import Image
import os

base = os.path.join("Tests", "images", "bmp")


class TestBmpReference(PillowTestCase):
    def get_files(self, d, ext=".bmp"):
        return [
            os.path.join(base, d, f)
            for f in os.listdir(os.path.join(base, d))
            if ext in f
        ]

    def test_bad(self):
        """ These shouldn't crash/dos, but they shouldn't return anything
        either """
        for f in self.get_files("b"):

            def open(f):
                try:
                    im = Image.open(f)
                    im.load()
                except Exception:  # as msg:
                    pass

            # Assert that there is no unclosed file warning
            self.assert_warning(None, open, f)

    def test_questionable(self):
        """ These shouldn't crash/dos, but it's not well defined that these
        are in spec """
        supported = [
            "pal8os2v2.bmp",
            "rgb24prof.bmp",
            "pal1p1.bmp",
            "pal8offs.bmp",
            "rgb24lprof.bmp",
            "rgb32fakealpha.bmp",
            "rgb24largepal.bmp",
            "pal8os2sp.bmp",
            "rgb32bf-xbgr.bmp",
        ]
        for f in self.get_files("q"):
            try:
                im = Image.open(f)
                im.load()
                if os.path.basename(f) not in supported:
                    print("Please add %s to the partially supported bmp specs." % f)
            except Exception:  # as msg:
                if os.path.basename(f) in supported:
                    raise

    def test_good(self):
        """ These should all work. There's a set of target files in the
        html directory that we can compare against. """

        # Target files, if they're not just replacing the extension
        file_map = {
            "pal1wb.bmp": "pal1.png",
            "pal4rle.bmp": "pal4.png",
            "pal8-0.bmp": "pal8.png",
            "pal8rle.bmp": "pal8.png",
            "pal8topdown.bmp": "pal8.png",
            "pal8nonsquare.bmp": "pal8nonsquare-v.png",
            "pal8os2.bmp": "pal8.png",
            "pal8os2sp.bmp": "pal8.png",
            "pal8os2v2.bmp": "pal8.png",
            "pal8os2v2-16.bmp": "pal8.png",
            "pal8v4.bmp": "pal8.png",
            "pal8v5.bmp": "pal8.png",
            "rgb16-565pal.bmp": "rgb16-565.png",
            "rgb24pal.bmp": "rgb24.png",
            "rgb32.bmp": "rgb24.png",
            "rgb32bf.bmp": "rgb24.png",
        }

        def get_compare(f):
            name = os.path.split(f)[1]
            if name in file_map:
                return os.path.join(base, "html", file_map[name])
            name = os.path.splitext(name)[0]
            return os.path.join(base, "html", "%s.png" % name)

        for f in self.get_files("g"):
            try:
                im = Image.open(f)
                im.load()
                compare = Image.open(get_compare(f))
                compare.load()
                if im.mode == "P":
                    # assert image similar doesn't really work
                    # with paletized image, since the palette might
                    # be differently ordered for an equivalent image.
                    im = im.convert("RGBA")
                    compare = im.convert("RGBA")
                self.assert_image_similar(im, compare, 5)

            except Exception as msg:
                # there are three here that are unsupported:
                unsupported = (
                    os.path.join(base, "g", "rgb32bf.bmp"),
                    os.path.join(base, "g", "pal8rle.bmp"),
                    os.path.join(base, "g", "pal4rle.bmp"),
                )
                if f not in unsupported:
                    self.fail("Unsupported Image %s: %s" % (f, msg))



from .helper import PillowTestCase, hopper
from array import array

import sys

from PIL import Image


class TestImagePutData(PillowTestCase):
    def test_sanity(self):

        im1 = hopper()

        data = list(im1.getdata())

        im2 = Image.new(im1.mode, im1.size, 0)
        im2.putdata(data)

        self.assert_image_equal(im1, im2)

        # readonly
        im2 = Image.new(im1.mode, im2.size, 0)
        im2.readonly = 1
        im2.putdata(data)

        self.assertFalse(im2.readonly)
        self.assert_image_equal(im1, im2)

    def test_long_integers(self):
        # see bug-200802-systemerror
        def put(value):
            im = Image.new("RGBA", (1, 1))
            im.putdata([value])
            return im.getpixel((0, 0))

        self.assertEqual(put(0xFFFFFFFF), (255, 255, 255, 255))
        self.assertEqual(put(0xFFFFFFFF), (255, 255, 255, 255))
        self.assertEqual(put(-1), (255, 255, 255, 255))
        self.assertEqual(put(-1), (255, 255, 255, 255))
        if sys.maxsize > 2 ** 32:
            self.assertEqual(put(sys.maxsize), (255, 255, 255, 255))
        else:
            self.assertEqual(put(sys.maxsize), (255, 255, 255, 127))

    def test_pypy_performance(self):
        im = Image.new("L", (256, 256))
        im.putdata(list(range(256)) * 256)

    def test_mode_i(self):
        src = hopper("L")
        data = list(src.getdata())
        im = Image.new("I", src.size, 0)
        im.putdata(data, 2, 256)

        target = [2 * elt + 256 for elt in data]
        self.assertEqual(list(im.getdata()), target)

    def test_mode_F(self):
        src = hopper("L")
        data = list(src.getdata())
        im = Image.new("F", src.size, 0)
        im.putdata(data, 2.0, 256.0)

        target = [2.0 * float(elt) + 256.0 for elt in data]
        self.assertEqual(list(im.getdata()), target)

    def test_array_B(self):
        # shouldn't segfault
        # see https://github.com/python-pillow/Pillow/issues/1008

        arr = array("B", [0]) * 15000
        im = Image.new("L", (150, 100))
        im.putdata(arr)

        self.assertEqual(len(im.getdata()), len(arr))

    def test_array_F(self):
        # shouldn't segfault
        # see https://github.com/python-pillow/Pillow/issues/1008

        im = Image.new("F", (150, 100))
        arr = array("f", [0.0]) * 15000
        im.putdata(arr)

        self.assertEqual(len(im.getdata()), len(arr))



from __future__ import print_function
from .helper import unittest, PillowTestCase

from PIL import Image

import locale

# ref https://github.com/python-pillow/Pillow/issues/272
# on windows, in polish locale:

# import locale
# print(locale.setlocale(locale.LC_ALL, 'polish'))
# import string
# print(len(string.whitespace))
# print(ord(string.whitespace[6]))

# Polish_Poland.1250
# 7
# 160

# one of string.whitespace is not freely convertable into ascii.

path = "Tests/images/hopper.jpg"


class TestLocale(PillowTestCase):
    def test_sanity(self):
        Image.open(path)
        try:
            locale.setlocale(locale.LC_ALL, "polish")
        except locale.Error:
            unittest.skip("Polish locale not available")
        Image.open(path)



from .helper import PillowTestCase, hopper
from PIL import Image


class TestImageThumbnail(PillowTestCase):
    def test_sanity(self):

        im = hopper()
        im.thumbnail((100, 100))

        self.assert_image(im, im.mode, (100, 100))

    def test_aspect(self):

        im = hopper()
        im.thumbnail((100, 100))
        self.assert_image(im, im.mode, (100, 100))

        im = hopper().resize((128, 256))
        im.thumbnail((100, 100))
        self.assert_image(im, im.mode, (50, 100))

        im = hopper().resize((128, 256))
        im.thumbnail((50, 100))
        self.assert_image(im, im.mode, (50, 100))

        im = hopper().resize((256, 128))
        im.thumbnail((100, 100))
        self.assert_image(im, im.mode, (100, 50))

        im = hopper().resize((256, 128))
        im.thumbnail((100, 50))
        self.assert_image(im, im.mode, (100, 50))

        im = hopper().resize((128, 128))
        im.thumbnail((100, 100))
        self.assert_image(im, im.mode, (100, 100))

    def test_no_resize(self):
        # Check that draft() can resize the image to the destination size
        im = Image.open("Tests/images/hopper.jpg")
        im.draft(None, (64, 64))
        self.assertEqual(im.size, (64, 64))

        # Test thumbnail(), where only draft() is necessary to resize the image
        im = Image.open("Tests/images/hopper.jpg")
        im.thumbnail((64, 64))
        self.assert_image(im, im.mode, (64, 64))



from .helper import PillowTestCase, hopper

from PIL import Image
from PIL._util import py3

import colorsys
import itertools


class TestFormatHSV(PillowTestCase):
    def int_to_float(self, i):
        return float(i) / 255.0

    def str_to_float(self, i):
        return float(ord(i)) / 255.0

    def tuple_to_ints(self, tp):
        x, y, z = tp
        return int(x * 255.0), int(y * 255.0), int(z * 255.0)

    def test_sanity(self):
        Image.new("HSV", (100, 100))

    def wedge(self):
        w = Image._wedge()
        w90 = w.rotate(90)

        (px, h) = w.size

        r = Image.new("L", (px * 3, h))
        g = r.copy()
        b = r.copy()

        r.paste(w, (0, 0))
        r.paste(w90, (px, 0))

        g.paste(w90, (0, 0))
        g.paste(w, (2 * px, 0))

        b.paste(w, (px, 0))
        b.paste(w90, (2 * px, 0))

        img = Image.merge("RGB", (r, g, b))

        return img

    def to_xxx_colorsys(self, im, func, mode):
        # convert the hard way using the library colorsys routines.

        (r, g, b) = im.split()

        if py3:
            conv_func = self.int_to_float
        else:
            conv_func = self.str_to_float

        if hasattr(itertools, "izip"):
            iter_helper = itertools.izip
        else:
            iter_helper = itertools.zip_longest

        converted = [
            self.tuple_to_ints(func(conv_func(_r), conv_func(_g), conv_func(_b)))
            for (_r, _g, _b) in iter_helper(r.tobytes(), g.tobytes(), b.tobytes())
        ]

        if py3:
            new_bytes = b"".join(
                bytes(chr(h) + chr(s) + chr(v), "latin-1") for (h, s, v) in converted
            )
        else:
            new_bytes = b"".join(chr(h) + chr(s) + chr(v) for (h, s, v) in converted)

        hsv = Image.frombytes(mode, r.size, new_bytes)

        return hsv

    def to_hsv_colorsys(self, im):
        return self.to_xxx_colorsys(im, colorsys.rgb_to_hsv, "HSV")

    def to_rgb_colorsys(self, im):
        return self.to_xxx_colorsys(im, colorsys.hsv_to_rgb, "RGB")

    def test_wedge(self):
        src = self.wedge().resize((3 * 32, 32), Image.BILINEAR)
        im = src.convert("HSV")
        comparable = self.to_hsv_colorsys(src)

        self.assert_image_similar(
            im.getchannel(0), comparable.getchannel(0), 1, "Hue conversion is wrong"
        )
        self.assert_image_similar(
            im.getchannel(1),
            comparable.getchannel(1),
            1,
            "Saturation conversion is wrong",
        )
        self.assert_image_similar(
            im.getchannel(2), comparable.getchannel(2), 1, "Value conversion is wrong"
        )

        comparable = src
        im = im.convert("RGB")

        self.assert_image_similar(
            im.getchannel(0), comparable.getchannel(0), 3, "R conversion is wrong"
        )
        self.assert_image_similar(
            im.getchannel(1), comparable.getchannel(1), 3, "G conversion is wrong"
        )
        self.assert_image_similar(
            im.getchannel(2), comparable.getchannel(2), 3, "B conversion is wrong"
        )

    def test_convert(self):
        im = hopper("RGB").convert("HSV")
        comparable = self.to_hsv_colorsys(hopper("RGB"))

        self.assert_image_similar(
            im.getchannel(0), comparable.getchannel(0), 1, "Hue conversion is wrong"
        )
        self.assert_image_similar(
            im.getchannel(1),
            comparable.getchannel(1),
            1,
            "Saturation conversion is wrong",
        )
        self.assert_image_similar(
            im.getchannel(2), comparable.getchannel(2), 1, "Value conversion is wrong"
        )

    def test_hsv_to_rgb(self):
        comparable = self.to_hsv_colorsys(hopper("RGB"))
        converted = comparable.convert("RGB")
        comparable = self.to_rgb_colorsys(comparable)

        self.assert_image_similar(
            converted.getchannel(0),
            comparable.getchannel(0),
            3,
            "R conversion is wrong",
        )
        self.assert_image_similar(
            converted.getchannel(1),
            comparable.getchannel(1),
            3,
            "G conversion is wrong",
        )
        self.assert_image_similar(
            converted.getchannel(2),
            comparable.getchannel(2),
            3,
            "B conversion is wrong",
        )



from .helper import PillowTestCase

from PIL import WalImageFile


class TestFileWal(PillowTestCase):
    def test_open(self):
        # Arrange
        TEST_FILE = "Tests/images/hopper.wal"

        # Act
        im = WalImageFile.open(TEST_FILE)

        # Assert
        self.assertEqual(im.format, "WAL")
        self.assertEqual(im.format_description, "Quake2 Texture")
        self.assertEqual(im.mode, "P")
        self.assertEqual(im.size, (128, 128))



from .helper import PillowTestCase

from PIL import Image

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False


class TestFileWebpMetadata(PillowTestCase):
    def setUp(self):
        if not HAVE_WEBP:
            self.skipTest("WebP support not installed")
            return

        if not _webp.HAVE_WEBPMUX:
            self.skipTest("WebPMux support not installed")

    def test_read_exif_metadata(self):

        file_path = "Tests/images/flower.webp"
        image = Image.open(file_path)

        self.assertEqual(image.format, "WEBP")
        exif_data = image.info.get("exif", None)
        self.assertTrue(exif_data)

        exif = image._getexif()

        # camera make
        self.assertEqual(exif[271], "Canon")

        jpeg_image = Image.open("Tests/images/flower.jpg")
        expected_exif = jpeg_image.info["exif"]

        self.assertEqual(exif_data, expected_exif)

    def test_write_exif_metadata(self):
        from io import BytesIO

        file_path = "Tests/images/flower.jpg"
        image = Image.open(file_path)
        expected_exif = image.info["exif"]

        test_buffer = BytesIO()

        image.save(test_buffer, "webp", exif=expected_exif)

        test_buffer.seek(0)
        webp_image = Image.open(test_buffer)

        webp_exif = webp_image.info.get("exif", None)
        self.assertTrue(webp_exif)
        if webp_exif:
            self.assertEqual(webp_exif, expected_exif, "WebP EXIF didn't match")

    def test_read_icc_profile(self):

        file_path = "Tests/images/flower2.webp"
        image = Image.open(file_path)

        self.assertEqual(image.format, "WEBP")
        self.assertTrue(image.info.get("icc_profile", None))

        icc = image.info["icc_profile"]

        jpeg_image = Image.open("Tests/images/flower2.jpg")
        expected_icc = jpeg_image.info["icc_profile"]

        self.assertEqual(icc, expected_icc)

    def test_write_icc_metadata(self):
        from io import BytesIO

        file_path = "Tests/images/flower2.jpg"
        image = Image.open(file_path)
        expected_icc_profile = image.info["icc_profile"]

        test_buffer = BytesIO()

        image.save(test_buffer, "webp", icc_profile=expected_icc_profile)

        test_buffer.seek(0)
        webp_image = Image.open(test_buffer)

        webp_icc_profile = webp_image.info.get("icc_profile", None)

        self.assertTrue(webp_icc_profile)
        if webp_icc_profile:
            self.assertEqual(
                webp_icc_profile, expected_icc_profile, "Webp ICC didn't match"
            )

    def test_read_no_exif(self):
        from io import BytesIO

        file_path = "Tests/images/flower.jpg"
        image = Image.open(file_path)
        self.assertIn("exif", image.info)

        test_buffer = BytesIO()

        image.save(test_buffer, "webp")

        test_buffer.seek(0)
        webp_image = Image.open(test_buffer)

        self.assertFalse(webp_image._getexif())

    def test_write_animated_metadata(self):
        if not _webp.HAVE_WEBPANIM:
            self.skipTest("WebP animation support not available")

        iccp_data = "<iccp_data>".encode("utf-8")
        exif_data = "<exif_data>".encode("utf-8")
        xmp_data = "<xmp_data>".encode("utf-8")

        temp_file = self.tempfile("temp.webp")
        frame1 = Image.open("Tests/images/anim_frame1.webp")
        frame2 = Image.open("Tests/images/anim_frame2.webp")
        frame1.save(
            temp_file,
            save_all=True,
            append_images=[frame2, frame1, frame2],
            icc_profile=iccp_data,
            exif=exif_data,
            xmp=xmp_data,
        )

        image = Image.open(temp_file)
        self.assertIn("icc_profile", image.info)
        self.assertIn("exif", image.info)
        self.assertIn("xmp", image.info)
        self.assertEqual(iccp_data, image.info.get("icc_profile", None))
        self.assertEqual(exif_data, image.info.get("exif", None))
        self.assertEqual(xmp_data, image.info.get("xmp", None))



from .helper import PillowTestCase, hopper

from PIL import Image


class TestImageSplit(PillowTestCase):
    def test_split(self):
        def split(mode):
            layers = hopper(mode).split()
            return [(i.mode, i.size[0], i.size[1]) for i in layers]

        self.assertEqual(split("1"), [("1", 128, 128)])
        self.assertEqual(split("L"), [("L", 128, 128)])
        self.assertEqual(split("I"), [("I", 128, 128)])
        self.assertEqual(split("F"), [("F", 128, 128)])
        self.assertEqual(split("P"), [("P", 128, 128)])
        self.assertEqual(
            split("RGB"), [("L", 128, 128), ("L", 128, 128), ("L", 128, 128)]
        )
        self.assertEqual(
            split("RGBA"),
            [("L", 128, 128), ("L", 128, 128), ("L", 128, 128), ("L", 128, 128)],
        )
        self.assertEqual(
            split("CMYK"),
            [("L", 128, 128), ("L", 128, 128), ("L", 128, 128), ("L", 128, 128)],
        )
        self.assertEqual(
            split("YCbCr"), [("L", 128, 128), ("L", 128, 128), ("L", 128, 128)]
        )

    def test_split_merge(self):
        def split_merge(mode):
            return Image.merge(mode, hopper(mode).split())

        self.assert_image_equal(hopper("1"), split_merge("1"))
        self.assert_image_equal(hopper("L"), split_merge("L"))
        self.assert_image_equal(hopper("I"), split_merge("I"))
        self.assert_image_equal(hopper("F"), split_merge("F"))
        self.assert_image_equal(hopper("P"), split_merge("P"))
        self.assert_image_equal(hopper("RGB"), split_merge("RGB"))
        self.assert_image_equal(hopper("RGBA"), split_merge("RGBA"))
        self.assert_image_equal(hopper("CMYK"), split_merge("CMYK"))
        self.assert_image_equal(hopper("YCbCr"), split_merge("YCbCr"))

    def test_split_open(self):
        codecs = dir(Image.core)

        if "zip_encoder" in codecs:
            test_file = self.tempfile("temp.png")
        else:
            test_file = self.tempfile("temp.pcx")

        def split_open(mode):
            hopper(mode).save(test_file)
            im = Image.open(test_file)
            return len(im.split())

        self.assertEqual(split_open("1"), 1)
        self.assertEqual(split_open("L"), 1)
        self.assertEqual(split_open("P"), 1)
        self.assertEqual(split_open("RGB"), 3)
        if "zip_encoder" in codecs:
            self.assertEqual(split_open("RGBA"), 4)



from .helper import PillowTestCase, hopper
from .test_imageqt import PillowQtTestCase

from PIL import ImageQt, Image


if ImageQt.qt_is_installed:
    from PIL.ImageQt import QImage

    try:
        from PyQt5 import QtGui
        from PyQt5.QtWidgets import QWidget, QHBoxLayout, QLabel, QApplication

        QT_VERSION = 5
    except (ImportError, RuntimeError):
        try:
            from PySide2 import QtGui
            from PySide2.QtWidgets import QWidget, QHBoxLayout, QLabel, QApplication

            QT_VERSION = 5
        except (ImportError, RuntimeError):
            try:
                from PyQt4 import QtGui
                from PyQt4.QtGui import QWidget, QHBoxLayout, QLabel, QApplication

                QT_VERSION = 4
            except (ImportError, RuntimeError):
                from PySide import QtGui
                from PySide.QtGui import QWidget, QHBoxLayout, QLabel, QApplication

                QT_VERSION = 4


class TestToQImage(PillowQtTestCase, PillowTestCase):
    def test_sanity(self):
        for mode in ("RGB", "RGBA", "L", "P", "1"):
            src = hopper(mode)
            data = ImageQt.toqimage(src)

            self.assertIsInstance(data, QImage)
            self.assertFalse(data.isNull())

            # reload directly from the qimage
            rt = ImageQt.fromqimage(data)
            if mode in ("L", "P", "1"):
                self.assert_image_equal(rt, src.convert("RGB"))
            else:
                self.assert_image_equal(rt, src)

            if mode == "1":
                # BW appears to not save correctly on QT4 and QT5
                # kicks out errors on console:
                #     libpng warning: Invalid color type/bit depth combination
                #                     in IHDR
                #     libpng error: Invalid IHDR data
                continue

            # Test saving the file
            tempfile = self.tempfile("temp_{}.png".format(mode))
            data.save(tempfile)

            # Check that it actually worked.
            reloaded = Image.open(tempfile)
            # Gray images appear to come back in palette mode.
            # They're roughly equivalent
            if QT_VERSION == 4 and mode == "L":
                src = src.convert("P")
            self.assert_image_equal(reloaded, src)

    def test_segfault(self):
        app = QApplication([])
        ex = Example()
        assert app  # Silence warning
        assert ex  # Silence warning


if ImageQt.qt_is_installed:

    class Example(QWidget):
        def __init__(self):
            super(Example, self).__init__()

            img = hopper().resize((1000, 1000))

            qimage = ImageQt.ImageQt(img)

            pixmap1 = QtGui.QPixmap.fromImage(qimage)

            QHBoxLayout(self)  # hbox

            lbl = QLabel(self)
            # Segfault in the problem
            lbl.setPixmap(pixmap1.copy())



from .helper import PillowTestCase, hopper

from PIL import Image
from PIL import ImageOps

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False


class TestImageOps(PillowTestCase):
    class Deformer(object):
        def getmesh(self, im):
            x, y = im.size
            return [((0, 0, x, y), (0, 0, x, 0, x, y, y, 0))]

    deformer = Deformer()

    def test_sanity(self):

        ImageOps.autocontrast(hopper("L"))
        ImageOps.autocontrast(hopper("RGB"))

        ImageOps.autocontrast(hopper("L"), cutoff=10)
        ImageOps.autocontrast(hopper("L"), ignore=[0, 255])

        ImageOps.colorize(hopper("L"), (0, 0, 0), (255, 255, 255))
        ImageOps.colorize(hopper("L"), "black", "white")

        ImageOps.pad(hopper("L"), (128, 128))
        ImageOps.pad(hopper("RGB"), (128, 128))

        ImageOps.crop(hopper("L"), 1)
        ImageOps.crop(hopper("RGB"), 1)

        ImageOps.deform(hopper("L"), self.deformer)
        ImageOps.deform(hopper("RGB"), self.deformer)

        ImageOps.equalize(hopper("L"))
        ImageOps.equalize(hopper("RGB"))

        ImageOps.expand(hopper("L"), 1)
        ImageOps.expand(hopper("RGB"), 1)
        ImageOps.expand(hopper("L"), 2, "blue")
        ImageOps.expand(hopper("RGB"), 2, "blue")

        ImageOps.fit(hopper("L"), (128, 128))
        ImageOps.fit(hopper("RGB"), (128, 128))

        ImageOps.flip(hopper("L"))
        ImageOps.flip(hopper("RGB"))

        ImageOps.grayscale(hopper("L"))
        ImageOps.grayscale(hopper("RGB"))

        ImageOps.invert(hopper("L"))
        ImageOps.invert(hopper("RGB"))

        ImageOps.mirror(hopper("L"))
        ImageOps.mirror(hopper("RGB"))

        ImageOps.posterize(hopper("L"), 4)
        ImageOps.posterize(hopper("RGB"), 4)

        ImageOps.solarize(hopper("L"))
        ImageOps.solarize(hopper("RGB"))

        ImageOps.exif_transpose(hopper("L"))
        ImageOps.exif_transpose(hopper("RGB"))

    def test_1pxfit(self):
        # Division by zero in equalize if image is 1 pixel high
        newimg = ImageOps.fit(hopper("RGB").resize((1, 1)), (35, 35))
        self.assertEqual(newimg.size, (35, 35))

        newimg = ImageOps.fit(hopper("RGB").resize((1, 100)), (35, 35))
        self.assertEqual(newimg.size, (35, 35))

        newimg = ImageOps.fit(hopper("RGB").resize((100, 1)), (35, 35))
        self.assertEqual(newimg.size, (35, 35))

    def test_pad(self):
        # Same ratio
        im = hopper()
        new_size = (im.width * 2, im.height * 2)
        new_im = ImageOps.pad(im, new_size)
        self.assertEqual(new_im.size, new_size)

        for label, color, new_size in [
            ("h", None, (im.width * 4, im.height * 2)),
            ("v", "#f00", (im.width * 2, im.height * 4)),
        ]:
            for i, centering in enumerate([(0, 0), (0.5, 0.5), (1, 1)]):
                new_im = ImageOps.pad(im, new_size, color=color, centering=centering)
                self.assertEqual(new_im.size, new_size)

                target = Image.open(
                    "Tests/images/imageops_pad_" + label + "_" + str(i) + ".jpg"
                )
                self.assert_image_similar(new_im, target, 6)

    def test_pil163(self):
        # Division by zero in equalize if < 255 pixels in image (@PIL163)

        i = hopper("RGB").resize((15, 16))

        ImageOps.equalize(i.convert("L"))
        ImageOps.equalize(i.convert("P"))
        ImageOps.equalize(i.convert("RGB"))

    def test_scale(self):
        # Test the scaling function
        i = hopper("L").resize((50, 50))

        with self.assertRaises(ValueError):
            ImageOps.scale(i, -1)

        newimg = ImageOps.scale(i, 1)
        self.assertEqual(newimg.size, (50, 50))

        newimg = ImageOps.scale(i, 2)
        self.assertEqual(newimg.size, (100, 100))

        newimg = ImageOps.scale(i, 0.5)
        self.assertEqual(newimg.size, (25, 25))

    def test_colorize_2color(self):
        # Test the colorizing function with 2-color functionality

        # Open test image (256px by 10px, black to white)
        im = Image.open("Tests/images/bw_gradient.png")
        im = im.convert("L")

        # Create image with original 2-color functionality
        im_test = ImageOps.colorize(im, "red", "green")

        # Test output image (2-color)
        left = (0, 1)
        middle = (127, 1)
        right = (255, 1)
        self.assert_tuple_approx_equal(
            im_test.getpixel(left),
            (255, 0, 0),
            threshold=1,
            msg="black test pixel incorrect",
        )
        self.assert_tuple_approx_equal(
            im_test.getpixel(middle),
            (127, 63, 0),
            threshold=1,
            msg="mid test pixel incorrect",
        )
        self.assert_tuple_approx_equal(
            im_test.getpixel(right),
            (0, 127, 0),
            threshold=1,
            msg="white test pixel incorrect",
        )

    def test_colorize_2color_offset(self):
        # Test the colorizing function with 2-color functionality and offset

        # Open test image (256px by 10px, black to white)
        im = Image.open("Tests/images/bw_gradient.png")
        im = im.convert("L")

        # Create image with original 2-color functionality with offsets
        im_test = ImageOps.colorize(
            im, black="red", white="green", blackpoint=50, whitepoint=100
        )

        # Test output image (2-color) with offsets
        left = (25, 1)
        middle = (75, 1)
        right = (125, 1)
        self.assert_tuple_approx_equal(
            im_test.getpixel(left),
            (255, 0, 0),
            threshold=1,
            msg="black test pixel incorrect",
        )
        self.assert_tuple_approx_equal(
            im_test.getpixel(middle),
            (127, 63, 0),
            threshold=1,
            msg="mid test pixel incorrect",
        )
        self.assert_tuple_approx_equal(
            im_test.getpixel(right),
            (0, 127, 0),
            threshold=1,
            msg="white test pixel incorrect",
        )

    def test_colorize_3color_offset(self):
        # Test the colorizing function with 3-color functionality and offset

        # Open test image (256px by 10px, black to white)
        im = Image.open("Tests/images/bw_gradient.png")
        im = im.convert("L")

        # Create image with new three color functionality with offsets
        im_test = ImageOps.colorize(
            im,
            black="red",
            white="green",
            mid="blue",
            blackpoint=50,
            whitepoint=200,
            midpoint=100,
        )

        # Test output image (3-color) with offsets
        left = (25, 1)
        left_middle = (75, 1)
        middle = (100, 1)
        right_middle = (150, 1)
        right = (225, 1)
        self.assert_tuple_approx_equal(
            im_test.getpixel(left),
            (255, 0, 0),
            threshold=1,
            msg="black test pixel incorrect",
        )
        self.assert_tuple_approx_equal(
            im_test.getpixel(left_middle),
            (127, 0, 127),
            threshold=1,
            msg="low-mid test pixel incorrect",
        )
        self.assert_tuple_approx_equal(
            im_test.getpixel(middle), (0, 0, 255), threshold=1, msg="mid incorrect"
        )
        self.assert_tuple_approx_equal(
            im_test.getpixel(right_middle),
            (0, 63, 127),
            threshold=1,
            msg="high-mid test pixel incorrect",
        )
        self.assert_tuple_approx_equal(
            im_test.getpixel(right),
            (0, 127, 0),
            threshold=1,
            msg="white test pixel incorrect",
        )

    def test_exif_transpose(self):
        exts = [".jpg"]
        if HAVE_WEBP and _webp.HAVE_WEBPANIM:
            exts.append(".webp")
        for ext in exts:
            base_im = Image.open("Tests/images/hopper" + ext)

            orientations = [base_im]
            for i in range(2, 9):
                im = Image.open("Tests/images/hopper_orientation_" + str(i) + ext)
                orientations.append(im)
            for i, orientation_im in enumerate(orientations):
                for im in [orientation_im, orientation_im.copy()]:  # ImageFile  # Image
                    if i == 0:
                        self.assertNotIn("exif", im.info)
                    else:
                        original_exif = im.info["exif"]
                    transposed_im = ImageOps.exif_transpose(im)
                    self.assert_image_similar(base_im, transposed_im, 17)
                    if i == 0:
                        self.assertNotIn("exif", im.info)
                    else:
                        self.assertNotEqual(transposed_im.info["exif"], original_exif)

                        self.assertNotIn(0x0112, transposed_im.getexif())

                    # Repeat the operation, to test that it does not keep transposing
                    transposed_im2 = ImageOps.exif_transpose(transposed_im)
                    self.assert_image_equal(transposed_im2, transposed_im)



#!/usr/bin/env python

from __future__ import division
from .helper import unittest, PillowTestCase
import sys
from PIL import Image

min_iterations = 100
max_iterations = 10000


@unittest.skipIf(sys.platform.startswith("win32"), "requires Unix or macOS")
class TestImagingLeaks(PillowTestCase):
    def _get_mem_usage(self):
        from resource import getpagesize, getrusage, RUSAGE_SELF

        mem = getrusage(RUSAGE_SELF).ru_maxrss
        return mem * getpagesize() / 1024 / 1024

    def _test_leak(self, min_iterations, max_iterations, fn, *args, **kwargs):
        mem_limit = None
        for i in range(max_iterations):
            fn(*args, **kwargs)
            mem = self._get_mem_usage()
            if i < min_iterations:
                mem_limit = mem + 1
                continue
            msg = "memory usage limit exceeded after %d iterations" % (i + 1)
            self.assertLessEqual(mem, mem_limit, msg)

    def test_leak_putdata(self):
        im = Image.new("RGB", (25, 25))
        self._test_leak(min_iterations, max_iterations, im.putdata, im.getdata())

    def test_leak_getlist(self):
        im = Image.new("P", (25, 25))
        self._test_leak(
            min_iterations,
            max_iterations,
            # Pass a new list at each iteration.
            lambda: im.point(range(256)),
        )


if __name__ == "__main__":
    unittest.main()



import logging
from io import BytesIO
import sys

from .helper import unittest, PillowTestCase, hopper

from PIL import Image, TiffImagePlugin, features
from PIL._util import py3
from PIL.TiffImagePlugin import X_RESOLUTION, Y_RESOLUTION, RESOLUTION_UNIT

logger = logging.getLogger(__name__)


class TestFileTiff(PillowTestCase):
    def test_sanity(self):

        filename = self.tempfile("temp.tif")

        hopper("RGB").save(filename)

        im = Image.open(filename)
        im.load()
        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.format, "TIFF")

        hopper("1").save(filename)
        Image.open(filename)

        hopper("L").save(filename)
        Image.open(filename)

        hopper("P").save(filename)
        Image.open(filename)

        hopper("RGB").save(filename)
        Image.open(filename)

        hopper("I").save(filename)
        Image.open(filename)

    def test_unclosed_file(self):
        def open():
            im = Image.open("Tests/images/multipage.tiff")
            im.load()

        self.assert_warning(None, open)

    def test_mac_tiff(self):
        # Read RGBa images from macOS [@PIL136]

        filename = "Tests/images/pil136.tiff"
        im = Image.open(filename)

        self.assertEqual(im.mode, "RGBA")
        self.assertEqual(im.size, (55, 43))
        self.assertEqual(im.tile, [("raw", (0, 0, 55, 43), 8, ("RGBa", 0, 1))])
        im.load()

        self.assert_image_similar_tofile(im, "Tests/images/pil136.png", 1)

    def test_wrong_bits_per_sample(self):
        im = Image.open("Tests/images/tiff_wrong_bits_per_sample.tiff")

        self.assertEqual(im.mode, "RGBA")
        self.assertEqual(im.size, (52, 53))
        self.assertEqual(im.tile, [("raw", (0, 0, 52, 53), 160, ("RGBA", 0, 1))])
        im.load()

    def test_set_legacy_api(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        with self.assertRaises(Exception) as e:
            ifd.legacy_api = None
        self.assertEqual(str(e.exception), "Not allowing setting of legacy api")

    def test_size(self):
        filename = "Tests/images/pil168.tif"
        im = Image.open(filename)

        def set_size():
            im.size = (256, 256)

        self.assert_warning(DeprecationWarning, set_size)

    def test_xyres_tiff(self):
        filename = "Tests/images/pil168.tif"
        im = Image.open(filename)

        # legacy api
        self.assertIsInstance(im.tag[X_RESOLUTION][0], tuple)
        self.assertIsInstance(im.tag[Y_RESOLUTION][0], tuple)

        # v2 api
        self.assertIsInstance(im.tag_v2[X_RESOLUTION], TiffImagePlugin.IFDRational)
        self.assertIsInstance(im.tag_v2[Y_RESOLUTION], TiffImagePlugin.IFDRational)

        self.assertEqual(im.info["dpi"], (72.0, 72.0))

    def test_xyres_fallback_tiff(self):
        filename = "Tests/images/compression.tif"
        im = Image.open(filename)

        # v2 api
        self.assertIsInstance(im.tag_v2[X_RESOLUTION], TiffImagePlugin.IFDRational)
        self.assertIsInstance(im.tag_v2[Y_RESOLUTION], TiffImagePlugin.IFDRational)
        self.assertRaises(KeyError, lambda: im.tag_v2[RESOLUTION_UNIT])

        # Legacy.
        self.assertEqual(im.info["resolution"], (100.0, 100.0))
        # Fallback "inch".
        self.assertEqual(im.info["dpi"], (100.0, 100.0))

    def test_int_resolution(self):
        filename = "Tests/images/pil168.tif"
        im = Image.open(filename)

        # Try to read a file where X,Y_RESOLUTION are ints
        im.tag_v2[X_RESOLUTION] = 71
        im.tag_v2[Y_RESOLUTION] = 71
        im._setup()
        self.assertEqual(im.info["dpi"], (71.0, 71.0))

    def test_load_dpi_rounding(self):
        for resolutionUnit, dpi in ((None, (72, 73)), (2, (72, 73)), (3, (183, 185))):
            im = Image.open(
                "Tests/images/hopper_roundDown_" + str(resolutionUnit) + ".tif"
            )
            self.assertEqual(im.tag_v2.get(RESOLUTION_UNIT), resolutionUnit)
            self.assertEqual(im.info["dpi"], (dpi[0], dpi[0]))

            im = Image.open(
                "Tests/images/hopper_roundUp_" + str(resolutionUnit) + ".tif"
            )
            self.assertEqual(im.tag_v2.get(RESOLUTION_UNIT), resolutionUnit)
            self.assertEqual(im.info["dpi"], (dpi[1], dpi[1]))

    def test_save_dpi_rounding(self):
        outfile = self.tempfile("temp.tif")
        im = Image.open("Tests/images/hopper.tif")

        for dpi in (72.2, 72.8):
            im.save(outfile, dpi=(dpi, dpi))

            reloaded = Image.open(outfile)
            reloaded.load()
            self.assertEqual((round(dpi), round(dpi)), reloaded.info["dpi"])

    def test_save_setting_missing_resolution(self):
        b = BytesIO()
        Image.open("Tests/images/10ct_32bit_128.tiff").save(
            b, format="tiff", resolution=123.45
        )
        im = Image.open(b)
        self.assertEqual(float(im.tag_v2[X_RESOLUTION]), 123.45)
        self.assertEqual(float(im.tag_v2[Y_RESOLUTION]), 123.45)

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, TiffImagePlugin.TiffImageFile, invalid_file)

        TiffImagePlugin.PREFIXES.append(b"\xff\xd8\xff\xe0")
        self.assertRaises(SyntaxError, TiffImagePlugin.TiffImageFile, invalid_file)
        TiffImagePlugin.PREFIXES.pop()

    def test_bad_exif(self):
        i = Image.open("Tests/images/hopper_bad_exif.jpg")
        # Should not raise struct.error.
        self.assert_warning(UserWarning, i._getexif)

    def test_save_rgba(self):
        im = hopper("RGBA")
        outfile = self.tempfile("temp.tif")
        im.save(outfile)

    def test_save_unsupported_mode(self):
        im = hopper("HSV")
        outfile = self.tempfile("temp.tif")
        self.assertRaises(IOError, im.save, outfile)

    def test_little_endian(self):
        im = Image.open("Tests/images/16bit.cropped.tif")
        self.assertEqual(im.getpixel((0, 0)), 480)
        self.assertEqual(im.mode, "I;16")

        b = im.tobytes()
        # Bytes are in image native order (little endian)
        if py3:
            self.assertEqual(b[0], ord(b"\xe0"))
            self.assertEqual(b[1], ord(b"\x01"))
        else:
            self.assertEqual(b[0], b"\xe0")
            self.assertEqual(b[1], b"\x01")

    def test_big_endian(self):
        im = Image.open("Tests/images/16bit.MM.cropped.tif")
        self.assertEqual(im.getpixel((0, 0)), 480)
        self.assertEqual(im.mode, "I;16B")

        b = im.tobytes()

        # Bytes are in image native order (big endian)
        if py3:
            self.assertEqual(b[0], ord(b"\x01"))
            self.assertEqual(b[1], ord(b"\xe0"))
        else:
            self.assertEqual(b[0], b"\x01")
            self.assertEqual(b[1], b"\xe0")

    def test_16bit_s(self):
        im = Image.open("Tests/images/16bit.s.tif")
        im.load()
        self.assertEqual(im.mode, "I")
        self.assertEqual(im.getpixel((0, 0)), 32767)
        self.assertEqual(im.getpixel((0, 1)), 0)

    def test_12bit_rawmode(self):
        """ Are we generating the same interpretation
        of the image as Imagemagick is? """

        im = Image.open("Tests/images/12bit.cropped.tif")

        # to make the target --
        # convert 12bit.cropped.tif -depth 16 tmp.tif
        # convert tmp.tif -evaluate RightShift 4 12in16bit2.tif
        # imagemagick will auto scale so that a 12bit FFF is 16bit FFF0,
        # so we need to unshift so that the integer values are the same.

        self.assert_image_equal_tofile(im, "Tests/images/12in16bit.tif")

    def test_32bit_float(self):
        # Issue 614, specific 32-bit float format
        path = "Tests/images/10ct_32bit_128.tiff"
        im = Image.open(path)
        im.load()

        self.assertEqual(im.getpixel((0, 0)), -0.4526388943195343)
        self.assertEqual(im.getextrema(), (-3.140936851501465, 3.140684127807617))

    def test_unknown_pixel_mode(self):
        self.assertRaises(
            IOError, Image.open, "Tests/images/hopper_unknown_pixel_mode.tif"
        )

    def test_n_frames(self):
        for path, n_frames in [
            ["Tests/images/multipage-lastframe.tif", 1],
            ["Tests/images/multipage.tiff", 3],
        ]:
            im = Image.open(path)
            self.assertEqual(im.n_frames, n_frames)
            self.assertEqual(im.is_animated, n_frames != 1)

    def test_eoferror(self):
        im = Image.open("Tests/images/multipage-lastframe.tif")
        n_frames = im.n_frames

        # Test seeking past the last frame
        self.assertRaises(EOFError, im.seek, n_frames)
        self.assertLess(im.tell(), n_frames)

        # Test that seeking to the last frame does not raise an error
        im.seek(n_frames - 1)

    def test_multipage(self):
        # issue #862
        im = Image.open("Tests/images/multipage.tiff")
        # file is a multipage tiff: 10x10 green, 10x10 red, 20x20 blue

        im.seek(0)
        self.assertEqual(im.size, (10, 10))
        self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 128, 0))

        im.seek(1)
        im.load()
        self.assertEqual(im.size, (10, 10))
        self.assertEqual(im.convert("RGB").getpixel((0, 0)), (255, 0, 0))

        im.seek(0)
        im.load()
        self.assertEqual(im.size, (10, 10))
        self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 128, 0))

        im.seek(2)
        im.load()
        self.assertEqual(im.size, (20, 20))
        self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 0, 255))

    def test_multipage_last_frame(self):
        im = Image.open("Tests/images/multipage-lastframe.tif")
        im.load()
        self.assertEqual(im.size, (20, 20))
        self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 0, 255))

    def test___str__(self):
        filename = "Tests/images/pil136.tiff"
        im = Image.open(filename)

        # Act
        ret = str(im.ifd)

        # Assert
        self.assertIsInstance(ret, str)

    def test_dict(self):
        # Arrange
        filename = "Tests/images/pil136.tiff"
        im = Image.open(filename)

        # v2 interface
        v2_tags = {
            256: 55,
            257: 43,
            258: (8, 8, 8, 8),
            259: 1,
            262: 2,
            296: 2,
            273: (8,),
            338: (1,),
            277: 4,
            279: (9460,),
            282: 72.0,
            283: 72.0,
            284: 1,
        }
        self.assertEqual(dict(im.tag_v2), v2_tags)

        # legacy interface
        legacy_tags = {
            256: (55,),
            257: (43,),
            258: (8, 8, 8, 8),
            259: (1,),
            262: (2,),
            296: (2,),
            273: (8,),
            338: (1,),
            277: (4,),
            279: (9460,),
            282: ((720000, 10000),),
            283: ((720000, 10000),),
            284: (1,),
        }
        self.assertEqual(dict(im.tag), legacy_tags)

    def test__delitem__(self):
        filename = "Tests/images/pil136.tiff"
        im = Image.open(filename)
        len_before = len(dict(im.ifd))
        del im.ifd[256]
        len_after = len(dict(im.ifd))
        self.assertEqual(len_before, len_after + 1)

    def test_load_byte(self):
        for legacy_api in [False, True]:
            ifd = TiffImagePlugin.ImageFileDirectory_v2()
            data = b"abc"
            ret = ifd.load_byte(data, legacy_api)
            self.assertEqual(ret, b"abc")

    def test_load_string(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        data = b"abc\0"
        ret = ifd.load_string(data, False)
        self.assertEqual(ret, "abc")

    def test_load_float(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        data = b"abcdabcd"
        ret = ifd.load_float(data, False)
        self.assertEqual(ret, (1.6777999408082104e22, 1.6777999408082104e22))

    def test_load_double(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        data = b"abcdefghabcdefgh"
        ret = ifd.load_double(data, False)
        self.assertEqual(ret, (8.540883223036124e194, 8.540883223036124e194))

    def test_seek(self):
        filename = "Tests/images/pil136.tiff"
        im = Image.open(filename)
        im.seek(0)
        self.assertEqual(im.tell(), 0)

    def test_seek_eof(self):
        filename = "Tests/images/pil136.tiff"
        im = Image.open(filename)
        self.assertEqual(im.tell(), 0)
        self.assertRaises(EOFError, im.seek, -1)
        self.assertRaises(EOFError, im.seek, 1)

    def test__limit_rational_int(self):
        from PIL.TiffImagePlugin import _limit_rational

        value = 34
        ret = _limit_rational(value, 65536)
        self.assertEqual(ret, (34, 1))

    def test__limit_rational_float(self):
        from PIL.TiffImagePlugin import _limit_rational

        value = 22.3
        ret = _limit_rational(value, 65536)
        self.assertEqual(ret, (223, 10))

    def test_4bit(self):
        test_file = "Tests/images/hopper_gray_4bpp.tif"
        original = hopper("L")
        im = Image.open(test_file)
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.mode, "L")
        self.assert_image_similar(im, original, 7.3)

    def test_gray_semibyte_per_pixel(self):
        test_files = (
            (
                24.8,  # epsilon
                (  # group
                    "Tests/images/tiff_gray_2_4_bpp/hopper2.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2I.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2R.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2IR.tif",
                ),
            ),
            (
                7.3,  # epsilon
                (  # group
                    "Tests/images/tiff_gray_2_4_bpp/hopper4.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4I.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4R.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4IR.tif",
                ),
            ),
        )
        original = hopper("L")
        for epsilon, group in test_files:
            im = Image.open(group[0])
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.mode, "L")
            self.assert_image_similar(im, original, epsilon)
            for file in group[1:]:
                im2 = Image.open(file)
                self.assertEqual(im2.size, (128, 128))
                self.assertEqual(im2.mode, "L")
                self.assert_image_equal(im, im2)

    def test_with_underscores(self):
        kwargs = {"resolution_unit": "inch", "x_resolution": 72, "y_resolution": 36}
        filename = self.tempfile("temp.tif")
        hopper("RGB").save(filename, **kwargs)
        im = Image.open(filename)

        # legacy interface
        self.assertEqual(im.tag[X_RESOLUTION][0][0], 72)
        self.assertEqual(im.tag[Y_RESOLUTION][0][0], 36)

        # v2 interface
        self.assertEqual(im.tag_v2[X_RESOLUTION], 72)
        self.assertEqual(im.tag_v2[Y_RESOLUTION], 36)

    def test_roundtrip_tiff_uint16(self):
        # Test an image of all '0' values
        pixel_value = 0x1234
        infile = "Tests/images/uint16_1_4660.tif"
        im = Image.open(infile)
        self.assertEqual(im.getpixel((0, 0)), pixel_value)

        tmpfile = self.tempfile("temp.tif")
        im.save(tmpfile)

        reloaded = Image.open(tmpfile)

        self.assert_image_equal(im, reloaded)

    def test_strip_raw(self):
        infile = "Tests/images/tiff_strip_raw.tif"
        im = Image.open(infile)

        self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_strip_planar_raw(self):
        # gdal_translate -of GTiff -co INTERLEAVE=BAND \
        # tiff_strip_raw.tif tiff_strip_planar_raw.tiff
        infile = "Tests/images/tiff_strip_planar_raw.tif"
        im = Image.open(infile)

        self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_strip_planar_raw_with_overviews(self):
        # gdaladdo tiff_strip_planar_raw2.tif 2 4 8 16
        infile = "Tests/images/tiff_strip_planar_raw_with_overviews.tif"
        im = Image.open(infile)

        self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_tiled_planar_raw(self):
        # gdal_translate -of GTiff -co TILED=YES -co BLOCKXSIZE=32 \
        # -co BLOCKYSIZE=32 -co INTERLEAVE=BAND \
        # tiff_tiled_raw.tif tiff_tiled_planar_raw.tiff
        infile = "Tests/images/tiff_tiled_planar_raw.tif"
        im = Image.open(infile)

        self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_palette(self):
        for mode in ["P", "PA"]:
            outfile = self.tempfile("temp.tif")

            im = hopper(mode)
            im.save(outfile)

            reloaded = Image.open(outfile)
            self.assert_image_equal(im.convert("RGB"), reloaded.convert("RGB"))

    def test_tiff_save_all(self):
        import io
        import os

        mp = io.BytesIO()
        with Image.open("Tests/images/multipage.tiff") as im:
            im.save(mp, format="tiff", save_all=True)

        mp.seek(0, os.SEEK_SET)
        with Image.open(mp) as im:
            self.assertEqual(im.n_frames, 3)

        # Test appending images
        mp = io.BytesIO()
        im = Image.new("RGB", (100, 100), "#f00")
        ims = [Image.new("RGB", (100, 100), color) for color in ["#0f0", "#00f"]]
        im.copy().save(mp, format="TIFF", save_all=True, append_images=ims)

        mp.seek(0, os.SEEK_SET)
        reread = Image.open(mp)
        self.assertEqual(reread.n_frames, 3)

        # Test appending using a generator
        def imGenerator(ims):
            for im in ims:
                yield im

        mp = io.BytesIO()
        im.save(mp, format="TIFF", save_all=True, append_images=imGenerator(ims))

        mp.seek(0, os.SEEK_SET)
        reread = Image.open(mp)
        self.assertEqual(reread.n_frames, 3)

    def test_saving_icc_profile(self):
        # Tests saving TIFF with icc_profile set.
        # At the time of writing this will only work for non-compressed tiffs
        # as libtiff does not support embedded ICC profiles,
        # ImageFile._save(..) however does.
        im = Image.new("RGB", (1, 1))
        im.info["icc_profile"] = "Dummy value"

        # Try save-load round trip to make sure both handle icc_profile.
        tmpfile = self.tempfile("temp.tif")
        im.save(tmpfile, "TIFF", compression="raw")
        reloaded = Image.open(tmpfile)

        self.assertEqual(b"Dummy value", reloaded.info["icc_profile"])

    def test_close_on_load_exclusive(self):
        # similar to test_fd_leak, but runs on unixlike os
        tmpfile = self.tempfile("temp.tif")

        with Image.open("Tests/images/uint16_1_4660.tif") as im:
            im.save(tmpfile)

        im = Image.open(tmpfile)
        fp = im.fp
        self.assertFalse(fp.closed)
        im.load()
        self.assertTrue(fp.closed)

    def test_close_on_load_nonexclusive(self):
        tmpfile = self.tempfile("temp.tif")

        with Image.open("Tests/images/uint16_1_4660.tif") as im:
            im.save(tmpfile)

        with open(tmpfile, "rb") as f:
            im = Image.open(f)
            fp = im.fp
            self.assertFalse(fp.closed)
            im.load()
            self.assertFalse(fp.closed)

    @unittest.skipUnless(features.check("libtiff"), "libtiff not installed")
    def test_sampleformat_not_corrupted(self):
        # Assert that a TIFF image with SampleFormat=UINT tag is not corrupted
        # when saving to a new file.
        # Pillow 6.0 fails with "OSError: cannot identify image file".
        import base64

        tiff = BytesIO(
            base64.b64decode(
                b"SUkqAAgAAAAPAP4ABAABAAAAAAAAAAABBAABAAAAAQAAAAEBBAABAAAAAQAA"
                b"AAIBAwADAAAAwgAAAAMBAwABAAAACAAAAAYBAwABAAAAAgAAABEBBAABAAAA"
                b"4AAAABUBAwABAAAAAwAAABYBBAABAAAAAQAAABcBBAABAAAACwAAABoBBQAB"
                b"AAAAyAAAABsBBQABAAAA0AAAABwBAwABAAAAAQAAACgBAwABAAAAAQAAAFMB"
                b"AwADAAAA2AAAAAAAAAAIAAgACAABAAAAAQAAAAEAAAABAAAAAQABAAEAAAB4"
                b"nGNgYAAAAAMAAQ=="
            )
        )
        out = BytesIO()
        with Image.open(tiff) as im:
            im.save(out, format="tiff")
        out.seek(0)
        with Image.open(out) as im:
            im.load()


@unittest.skipUnless(sys.platform.startswith("win32"), "Windows only")
class TestFileTiffW32(PillowTestCase):
    def test_fd_leak(self):
        tmpfile = self.tempfile("temp.tif")
        import os

        # this is an mmaped file.
        with Image.open("Tests/images/uint16_1_4660.tif") as im:
            im.save(tmpfile)

        im = Image.open(tmpfile)
        fp = im.fp
        self.assertFalse(fp.closed)
        self.assertRaises(WindowsError, os.remove, tmpfile)
        im.load()
        self.assertTrue(fp.closed)

        # this closes the mmap
        im.close()

        # this should not fail, as load should have closed the file pointer,
        # and close should have closed the mmap
        os.remove(tmpfile)



from __future__ import division, print_function

from contextlib import contextmanager

from .helper import unittest, PillowTestCase, hopper
from PIL import Image, ImageDraw


class TestImagingResampleVulnerability(PillowTestCase):
    # see https://github.com/python-pillow/Pillow/issues/1710
    def test_overflow(self):
        im = hopper("L")
        xsize = 0x100000008 // 4
        ysize = 1000  # unimportant
        with self.assertRaises(MemoryError):
            # any resampling filter will do here
            im.im.resize((xsize, ysize), Image.BILINEAR)

    def test_invalid_size(self):
        im = hopper()

        # Should not crash
        im.resize((100, 100))

        with self.assertRaises(ValueError):
            im.resize((-100, 100))

        with self.assertRaises(ValueError):
            im.resize((100, -100))

    def test_modify_after_resizing(self):
        im = hopper("RGB")
        # get copy with same size
        copy = im.resize(im.size)
        # some in-place operation
        copy.paste("black", (0, 0, im.width // 2, im.height // 2))
        # image should be different
        self.assertNotEqual(im.tobytes(), copy.tobytes())


class TestImagingCoreResampleAccuracy(PillowTestCase):
    def make_case(self, mode, size, color):
        """Makes a sample image with two dark and two bright squares.
        For example:
        e0 e0 1f 1f
        e0 e0 1f 1f
        1f 1f e0 e0
        1f 1f e0 e0
        """
        case = Image.new("L", size, 255 - color)
        rectangle = ImageDraw.Draw(case).rectangle
        rectangle((0, 0, size[0] // 2 - 1, size[1] // 2 - 1), color)
        rectangle((size[0] // 2, size[1] // 2, size[0], size[1]), color)

        return Image.merge(mode, [case] * len(mode))

    def make_sample(self, data, size):
        """Restores a sample image from given data string which contains
        hex-encoded pixels from the top left fourth of a sample.
        """
        data = data.replace(" ", "")
        sample = Image.new("L", size)
        s_px = sample.load()
        w, h = size[0] // 2, size[1] // 2
        for y in range(h):
            for x in range(w):
                val = int(data[(y * w + x) * 2 : (y * w + x + 1) * 2], 16)
                s_px[x, y] = val
                s_px[size[0] - x - 1, size[1] - y - 1] = val
                s_px[x, size[1] - y - 1] = 255 - val
                s_px[size[0] - x - 1, y] = 255 - val
        return sample

    def check_case(self, case, sample):
        s_px = sample.load()
        c_px = case.load()
        for y in range(case.size[1]):
            for x in range(case.size[0]):
                if c_px[x, y] != s_px[x, y]:
                    message = "\nHave: \n{}\n\nExpected: \n{}".format(
                        self.serialize_image(case), self.serialize_image(sample)
                    )
                    self.assertEqual(s_px[x, y], c_px[x, y], message)

    def serialize_image(self, image):
        s_px = image.load()
        return "\n".join(
            " ".join("{:02x}".format(s_px[x, y]) for x in range(image.size[0]))
            for y in range(image.size[1])
        )

    def test_reduce_box(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (8, 8), 0xE1)
            case = case.resize((4, 4), Image.BOX)
            # fmt: off
            data = ("e1 e1"
                    "e1 e1")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (4, 4)))

    def test_reduce_bilinear(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (8, 8), 0xE1)
            case = case.resize((4, 4), Image.BILINEAR)
            # fmt: off
            data = ("e1 c9"
                    "c9 b7")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (4, 4)))

    def test_reduce_hamming(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (8, 8), 0xE1)
            case = case.resize((4, 4), Image.HAMMING)
            # fmt: off
            data = ("e1 da"
                    "da d3")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (4, 4)))

    def test_reduce_bicubic(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (12, 12), 0xE1)
            case = case.resize((6, 6), Image.BICUBIC)
            # fmt: off
            data = ("e1 e3 d4"
                    "e3 e5 d6"
                    "d4 d6 c9")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (6, 6)))

    def test_reduce_lanczos(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (16, 16), 0xE1)
            case = case.resize((8, 8), Image.LANCZOS)
            # fmt: off
            data = ("e1 e0 e4 d7"
                    "e0 df e3 d6"
                    "e4 e3 e7 da"
                    "d7 d6 d9 ce")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (8, 8)))

    def test_enlarge_box(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (2, 2), 0xE1)
            case = case.resize((4, 4), Image.BOX)
            # fmt: off
            data = ("e1 e1"
                    "e1 e1")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (4, 4)))

    def test_enlarge_bilinear(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (2, 2), 0xE1)
            case = case.resize((4, 4), Image.BILINEAR)
            # fmt: off
            data = ("e1 b0"
                    "b0 98")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (4, 4)))

    def test_enlarge_hamming(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (2, 2), 0xE1)
            case = case.resize((4, 4), Image.HAMMING)
            # fmt: off
            data = ("e1 d2"
                    "d2 c5")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (4, 4)))

    def test_enlarge_bicubic(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (4, 4), 0xE1)
            case = case.resize((8, 8), Image.BICUBIC)
            # fmt: off
            data = ("e1 e5 ee b9"
                    "e5 e9 f3 bc"
                    "ee f3 fd c1"
                    "b9 bc c1 a2")
            # fmt: on
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (8, 8)))

    def test_enlarge_lanczos(self):
        for mode in ["RGBX", "RGB", "La", "L"]:
            case = self.make_case(mode, (6, 6), 0xE1)
            case = case.resize((12, 12), Image.LANCZOS)
            data = (
                "e1 e0 db ed f5 b8"
                "e0 df da ec f3 b7"
                "db db d6 e7 ee b5"
                "ed ec e6 fb ff bf"
                "f5 f4 ee ff ff c4"
                "b8 b7 b4 bf c4 a0"
            )
            for channel in case.split():
                self.check_case(channel, self.make_sample(data, (12, 12)))


class CoreResampleConsistencyTest(PillowTestCase):
    def make_case(self, mode, fill):
        im = Image.new(mode, (512, 9), fill)
        return im.resize((9, 512), Image.LANCZOS), im.load()[0, 0]

    def run_case(self, case):
        channel, color = case
        px = channel.load()
        for x in range(channel.size[0]):
            for y in range(channel.size[1]):
                if px[x, y] != color:
                    message = "{} != {} for pixel {}".format(px[x, y], color, (x, y))
                    self.assertEqual(px[x, y], color, message)

    def test_8u(self):
        im, color = self.make_case("RGB", (0, 64, 255))
        r, g, b = im.split()
        self.run_case((r, color[0]))
        self.run_case((g, color[1]))
        self.run_case((b, color[2]))
        self.run_case(self.make_case("L", 12))

    def test_32i(self):
        self.run_case(self.make_case("I", 12))
        self.run_case(self.make_case("I", 0x7FFFFFFF))
        self.run_case(self.make_case("I", -12))
        self.run_case(self.make_case("I", -1 << 31))

    def test_32f(self):
        self.run_case(self.make_case("F", 1))
        self.run_case(self.make_case("F", 3.40282306074e38))
        self.run_case(self.make_case("F", 1.175494e-38))
        self.run_case(self.make_case("F", 1.192093e-07))


class CoreResampleAlphaCorrectTest(PillowTestCase):
    def make_levels_case(self, mode):
        i = Image.new(mode, (256, 16))
        px = i.load()
        for y in range(i.size[1]):
            for x in range(i.size[0]):
                pix = [x] * len(mode)
                pix[-1] = 255 - y * 16
                px[x, y] = tuple(pix)
        return i

    def run_levels_case(self, i):
        px = i.load()
        for y in range(i.size[1]):
            used_colors = {px[x, y][0] for x in range(i.size[0])}
            self.assertEqual(
                256,
                len(used_colors),
                "All colors should present in resized image. "
                "Only {} on {} line.".format(len(used_colors), y),
            )

    @unittest.skip("current implementation isn't precise enough")
    def test_levels_rgba(self):
        case = self.make_levels_case("RGBA")
        self.run_levels_case(case.resize((512, 32), Image.BOX))
        self.run_levels_case(case.resize((512, 32), Image.BILINEAR))
        self.run_levels_case(case.resize((512, 32), Image.HAMMING))
        self.run_levels_case(case.resize((512, 32), Image.BICUBIC))
        self.run_levels_case(case.resize((512, 32), Image.LANCZOS))

    @unittest.skip("current implementation isn't precise enough")
    def test_levels_la(self):
        case = self.make_levels_case("LA")
        self.run_levels_case(case.resize((512, 32), Image.BOX))
        self.run_levels_case(case.resize((512, 32), Image.BILINEAR))
        self.run_levels_case(case.resize((512, 32), Image.HAMMING))
        self.run_levels_case(case.resize((512, 32), Image.BICUBIC))
        self.run_levels_case(case.resize((512, 32), Image.LANCZOS))

    def make_dirty_case(self, mode, clean_pixel, dirty_pixel):
        i = Image.new(mode, (64, 64), dirty_pixel)
        px = i.load()
        xdiv4 = i.size[0] // 4
        ydiv4 = i.size[1] // 4
        for y in range(ydiv4 * 2):
            for x in range(xdiv4 * 2):
                px[x + xdiv4, y + ydiv4] = clean_pixel
        return i

    def run_dirty_case(self, i, clean_pixel):
        px = i.load()
        for y in range(i.size[1]):
            for x in range(i.size[0]):
                if px[x, y][-1] != 0 and px[x, y][:-1] != clean_pixel:
                    message = "pixel at ({}, {}) is differ:\n{}\n{}".format(
                        x, y, px[x, y], clean_pixel
                    )
                    self.assertEqual(px[x, y][:3], clean_pixel, message)

    def test_dirty_pixels_rgba(self):
        case = self.make_dirty_case("RGBA", (255, 255, 0, 128), (0, 0, 255, 0))
        self.run_dirty_case(case.resize((20, 20), Image.BOX), (255, 255, 0))
        self.run_dirty_case(case.resize((20, 20), Image.BILINEAR), (255, 255, 0))
        self.run_dirty_case(case.resize((20, 20), Image.HAMMING), (255, 255, 0))
        self.run_dirty_case(case.resize((20, 20), Image.BICUBIC), (255, 255, 0))
        self.run_dirty_case(case.resize((20, 20), Image.LANCZOS), (255, 255, 0))

    def test_dirty_pixels_la(self):
        case = self.make_dirty_case("LA", (255, 128), (0, 0))
        self.run_dirty_case(case.resize((20, 20), Image.BOX), (255,))
        self.run_dirty_case(case.resize((20, 20), Image.BILINEAR), (255,))
        self.run_dirty_case(case.resize((20, 20), Image.HAMMING), (255,))
        self.run_dirty_case(case.resize((20, 20), Image.BICUBIC), (255,))
        self.run_dirty_case(case.resize((20, 20), Image.LANCZOS), (255,))


class CoreResamplePassesTest(PillowTestCase):
    @contextmanager
    def count(self, diff):
        count = Image.core.get_stats()["new_count"]
        yield
        self.assertEqual(Image.core.get_stats()["new_count"] - count, diff)

    def test_horizontal(self):
        im = hopper("L")
        with self.count(1):
            im.resize((im.size[0] - 10, im.size[1]), Image.BILINEAR)

    def test_vertical(self):
        im = hopper("L")
        with self.count(1):
            im.resize((im.size[0], im.size[1] - 10), Image.BILINEAR)

    def test_both(self):
        im = hopper("L")
        with self.count(2):
            im.resize((im.size[0] - 10, im.size[1] - 10), Image.BILINEAR)

    def test_box_horizontal(self):
        im = hopper("L")
        box = (20, 0, im.size[0] - 20, im.size[1])
        with self.count(1):
            # the same size, but different box
            with_box = im.resize(im.size, Image.BILINEAR, box)
        with self.count(2):
            cropped = im.crop(box).resize(im.size, Image.BILINEAR)
        self.assert_image_similar(with_box, cropped, 0.1)

    def test_box_vertical(self):
        im = hopper("L")
        box = (0, 20, im.size[0], im.size[1] - 20)
        with self.count(1):
            # the same size, but different box
            with_box = im.resize(im.size, Image.BILINEAR, box)
        with self.count(2):
            cropped = im.crop(box).resize(im.size, Image.BILINEAR)
        self.assert_image_similar(with_box, cropped, 0.1)


class CoreResampleCoefficientsTest(PillowTestCase):
    def test_reduce(self):
        test_color = 254

        for size in range(400000, 400010, 2):
            i = Image.new("L", (size, 1), 0)
            draw = ImageDraw.Draw(i)
            draw.rectangle((0, 0, i.size[0] // 2 - 1, 0), test_color)

            px = i.resize((5, i.size[1]), Image.BICUBIC).load()
            if px[2, 0] != test_color // 2:
                self.assertEqual(test_color // 2, px[2, 0])

    def test_nonzero_coefficients(self):
        # regression test for the wrong coefficients calculation
        # due to bug https://github.com/python-pillow/Pillow/issues/2161
        im = Image.new("RGBA", (1280, 1280), (0x20, 0x40, 0x60, 0xFF))
        histogram = im.resize((256, 256), Image.BICUBIC).histogram()

        # first channel
        self.assertEqual(histogram[0x100 * 0 + 0x20], 0x10000)
        # second channel
        self.assertEqual(histogram[0x100 * 1 + 0x40], 0x10000)
        # third channel
        self.assertEqual(histogram[0x100 * 2 + 0x60], 0x10000)
        # fourth channel
        self.assertEqual(histogram[0x100 * 3 + 0xFF], 0x10000)


class CoreResampleBoxTest(PillowTestCase):
    def test_wrong_arguments(self):
        im = hopper()
        for resample in (
            Image.NEAREST,
            Image.BOX,
            Image.BILINEAR,
            Image.HAMMING,
            Image.BICUBIC,
            Image.LANCZOS,
        ):
            im.resize((32, 32), resample, (0, 0, im.width, im.height))
            im.resize((32, 32), resample, (20, 20, im.width, im.height))
            im.resize((32, 32), resample, (20, 20, 20, 100))
            im.resize((32, 32), resample, (20, 20, 100, 20))

            with self.assertRaisesRegex(TypeError, "must be sequence of length 4"):
                im.resize((32, 32), resample, (im.width, im.height))

            with self.assertRaisesRegex(ValueError, "can't be negative"):
                im.resize((32, 32), resample, (-20, 20, 100, 100))
            with self.assertRaisesRegex(ValueError, "can't be negative"):
                im.resize((32, 32), resample, (20, -20, 100, 100))

            with self.assertRaisesRegex(ValueError, "can't be empty"):
                im.resize((32, 32), resample, (20.1, 20, 20, 100))
            with self.assertRaisesRegex(ValueError, "can't be empty"):
                im.resize((32, 32), resample, (20, 20.1, 100, 20))
            with self.assertRaisesRegex(ValueError, "can't be empty"):
                im.resize((32, 32), resample, (20.1, 20.1, 20, 20))

            with self.assertRaisesRegex(ValueError, "can't exceed"):
                im.resize((32, 32), resample, (0, 0, im.width + 1, im.height))
            with self.assertRaisesRegex(ValueError, "can't exceed"):
                im.resize((32, 32), resample, (0, 0, im.width, im.height + 1))

    def resize_tiled(self, im, dst_size, xtiles, ytiles):
        def split_range(size, tiles):
            scale = size / tiles
            for i in range(tiles):
                yield (int(round(scale * i)), int(round(scale * (i + 1))))

        tiled = Image.new(im.mode, dst_size)
        scale = (im.size[0] / tiled.size[0], im.size[1] / tiled.size[1])

        for y0, y1 in split_range(dst_size[1], ytiles):
            for x0, x1 in split_range(dst_size[0], xtiles):
                box = (x0 * scale[0], y0 * scale[1], x1 * scale[0], y1 * scale[1])
                tile = im.resize((x1 - x0, y1 - y0), Image.BICUBIC, box)
                tiled.paste(tile, (x0, y0))
        return tiled

    def test_tiles(self):
        im = Image.open("Tests/images/flower.jpg")
        self.assertEqual(im.size, (480, 360))
        dst_size = (251, 188)
        reference = im.resize(dst_size, Image.BICUBIC)

        for tiles in [(1, 1), (3, 3), (9, 7), (100, 100)]:
            tiled = self.resize_tiled(im, dst_size, *tiles)
            self.assert_image_similar(reference, tiled, 0.01)

    def test_subsample(self):
        # This test shows advantages of the subpixel resizing
        # after supersampling (e.g. during JPEG decoding).
        im = Image.open("Tests/images/flower.jpg")
        self.assertEqual(im.size, (480, 360))
        dst_size = (48, 36)
        # Reference is cropped image resized to destination
        reference = im.crop((0, 0, 473, 353)).resize(dst_size, Image.BICUBIC)
        # Image.BOX emulates supersampling (480 / 8 = 60, 360 / 8 = 45)
        supersampled = im.resize((60, 45), Image.BOX)

        with_box = supersampled.resize(dst_size, Image.BICUBIC, (0, 0, 59.125, 44.125))
        without_box = supersampled.resize(dst_size, Image.BICUBIC)

        # error with box should be much smaller than without
        self.assert_image_similar(reference, with_box, 6)
        with self.assertRaisesRegex(AssertionError, r"difference 29\."):
            self.assert_image_similar(reference, without_box, 5)

    def test_formats(self):
        for resample in [Image.NEAREST, Image.BILINEAR]:
            for mode in ["RGB", "L", "RGBA", "LA", "I", ""]:
                im = hopper(mode)
                box = (20, 20, im.size[0] - 20, im.size[1] - 20)
                with_box = im.resize((32, 32), resample, box)
                cropped = im.crop(box).resize((32, 32), resample)
                self.assert_image_similar(cropped, with_box, 0.4)

    def test_passthrough(self):
        # When no resize is required
        im = hopper()

        for size, box in [
            ((40, 50), (0, 0, 40, 50)),
            ((40, 50), (0, 10, 40, 60)),
            ((40, 50), (10, 0, 50, 50)),
            ((40, 50), (10, 20, 50, 70)),
        ]:
            try:
                res = im.resize(size, Image.LANCZOS, box)
                self.assertEqual(res.size, size)
                self.assert_image_equal(res, im.crop(box))
            except AssertionError:
                print(">>>", size, box)
                raise

    def test_no_passthrough(self):
        # When resize is required
        im = hopper()

        for size, box in [
            ((40, 50), (0.4, 0.4, 40.4, 50.4)),
            ((40, 50), (0.4, 10.4, 40.4, 60.4)),
            ((40, 50), (10.4, 0.4, 50.4, 50.4)),
            ((40, 50), (10.4, 20.4, 50.4, 70.4)),
        ]:
            try:
                res = im.resize(size, Image.LANCZOS, box)
                self.assertEqual(res.size, size)
                with self.assertRaisesRegex(AssertionError, r"difference \d"):
                    # check that the difference at least that much
                    self.assert_image_similar(res, im.crop(box), 20)
            except AssertionError:
                print(">>>", size, box)
                raise

    def test_skip_horizontal(self):
        # Can skip resize for one dimension
        im = hopper()

        for flt in [Image.NEAREST, Image.BICUBIC]:
            for size, box in [
                ((40, 50), (0, 0, 40, 90)),
                ((40, 50), (0, 20, 40, 90)),
                ((40, 50), (10, 0, 50, 90)),
                ((40, 50), (10, 20, 50, 90)),
            ]:
                try:
                    res = im.resize(size, flt, box)
                    self.assertEqual(res.size, size)
                    # Borders should be slightly different
                    self.assert_image_similar(res, im.crop(box).resize(size, flt), 0.4)
                except AssertionError:
                    print(">>>", size, box, flt)
                    raise

    def test_skip_vertical(self):
        # Can skip resize for one dimension
        im = hopper()

        for flt in [Image.NEAREST, Image.BICUBIC]:
            for size, box in [
                ((40, 50), (0, 0, 90, 50)),
                ((40, 50), (20, 0, 90, 50)),
                ((40, 50), (0, 10, 90, 60)),
                ((40, 50), (20, 10, 90, 60)),
            ]:
                try:
                    res = im.resize(size, flt, box)
                    self.assertEqual(res.size, size)
                    # Borders should be slightly different
                    self.assert_image_similar(res, im.crop(box).resize(size, flt), 0.4)
                except AssertionError:
                    print(">>>", size, box, flt)
                    raise



from .helper import PillowTestCase, hopper

from PIL import Image


class TestImageGetBbox(PillowTestCase):
    def test_sanity(self):

        bbox = hopper().getbbox()
        self.assertIsInstance(bbox, tuple)

    def test_bbox(self):

        # 8-bit mode
        im = Image.new("L", (100, 100), 0)
        self.assertIsNone(im.getbbox())

        im.paste(255, (10, 25, 90, 75))
        self.assertEqual(im.getbbox(), (10, 25, 90, 75))

        im.paste(255, (25, 10, 75, 90))
        self.assertEqual(im.getbbox(), (10, 10, 90, 90))

        im.paste(255, (-10, -10, 110, 110))
        self.assertEqual(im.getbbox(), (0, 0, 100, 100))

        # 32-bit mode
        im = Image.new("RGB", (100, 100), 0)
        self.assertIsNone(im.getbbox())

        im.paste(255, (10, 25, 90, 75))
        self.assertEqual(im.getbbox(), (10, 25, 90, 75))

        im.paste(255, (25, 10, 75, 90))
        self.assertEqual(im.getbbox(), (10, 10, 90, 90))

        im.paste(255, (-10, -10, 110, 110))
        self.assertEqual(im.getbbox(), (0, 0, 100, 100))



from PIL import Image

from .helper import PillowTestCase


class TestFileBlp(PillowTestCase):
    def test_load_blp2_raw(self):
        im = Image.open("Tests/images/blp/blp2_raw.blp")
        target = Image.open("Tests/images/blp/blp2_raw.png")
        self.assert_image_equal(im, target)

    def test_load_blp2_dxt1(self):
        im = Image.open("Tests/images/blp/blp2_dxt1.blp")
        target = Image.open("Tests/images/blp/blp2_dxt1.png")
        self.assert_image_equal(im, target)

    def test_load_blp2_dxt1a(self):
        im = Image.open("Tests/images/blp/blp2_dxt1a.blp")
        target = Image.open("Tests/images/blp/blp2_dxt1a.png")
        self.assert_image_equal(im, target)



from .helper import PillowTestCase

from PIL.PdfParser import (
    IndirectObjectDef,
    IndirectReference,
    PdfBinary,
    PdfDict,
    PdfFormatError,
    PdfName,
    PdfParser,
    PdfStream,
    decode_text,
    encode_text,
    pdf_repr,
)
import time


class TestPdfParser(PillowTestCase):
    def test_text_encode_decode(self):
        self.assertEqual(encode_text("abc"), b"\xFE\xFF\x00a\x00b\x00c")
        self.assertEqual(decode_text(b"\xFE\xFF\x00a\x00b\x00c"), "abc")
        self.assertEqual(decode_text(b"abc"), "abc")
        self.assertEqual(decode_text(b"\x1B a \x1C"), u"\u02D9 a \u02DD")

    def test_indirect_refs(self):
        self.assertEqual(IndirectReference(1, 2), IndirectReference(1, 2))
        self.assertNotEqual(IndirectReference(1, 2), IndirectReference(1, 3))
        self.assertNotEqual(IndirectReference(1, 2), IndirectObjectDef(1, 2))
        self.assertNotEqual(IndirectReference(1, 2), (1, 2))
        self.assertEqual(IndirectObjectDef(1, 2), IndirectObjectDef(1, 2))
        self.assertNotEqual(IndirectObjectDef(1, 2), IndirectObjectDef(1, 3))
        self.assertNotEqual(IndirectObjectDef(1, 2), IndirectReference(1, 2))
        self.assertNotEqual(IndirectObjectDef(1, 2), (1, 2))

    def test_parsing(self):
        self.assertEqual(PdfParser.interpret_name(b"Name#23Hash"), b"Name#Hash")
        self.assertEqual(
            PdfParser.interpret_name(b"Name#23Hash", as_text=True), "Name#Hash"
        )
        self.assertEqual(
            PdfParser.get_value(b"1 2 R ", 0), (IndirectReference(1, 2), 5)
        )
        self.assertEqual(PdfParser.get_value(b"true[", 0), (True, 4))
        self.assertEqual(PdfParser.get_value(b"false%", 0), (False, 5))
        self.assertEqual(PdfParser.get_value(b"null<", 0), (None, 4))
        self.assertEqual(PdfParser.get_value(b"%cmt\n %cmt\n 123\n", 0), (123, 15))
        self.assertEqual(PdfParser.get_value(b"<901FA3>", 0), (b"\x90\x1F\xA3", 8))
        self.assertEqual(
            PdfParser.get_value(b"asd < 9 0 1 f A > qwe", 3), (b"\x90\x1F\xA0", 17)
        )
        self.assertEqual(PdfParser.get_value(b"(asd)", 0), (b"asd", 5))
        self.assertEqual(
            PdfParser.get_value(b"(asd(qwe)zxc)zzz(aaa)", 0), (b"asd(qwe)zxc", 13)
        )
        self.assertEqual(
            PdfParser.get_value(b"(Two \\\nwords.)", 0), (b"Two words.", 14)
        )
        self.assertEqual(PdfParser.get_value(b"(Two\nlines.)", 0), (b"Two\nlines.", 12))
        self.assertEqual(
            PdfParser.get_value(b"(Two\r\nlines.)", 0), (b"Two\nlines.", 13)
        )
        self.assertEqual(
            PdfParser.get_value(b"(Two\\nlines.)", 0), (b"Two\nlines.", 13)
        )
        self.assertEqual(PdfParser.get_value(b"(One\\(paren).", 0), (b"One(paren", 12))
        self.assertEqual(PdfParser.get_value(b"(One\\)paren).", 0), (b"One)paren", 12))
        self.assertEqual(PdfParser.get_value(b"(\\0053)", 0), (b"\x053", 7))
        self.assertEqual(PdfParser.get_value(b"(\\053)", 0), (b"\x2B", 6))
        self.assertEqual(PdfParser.get_value(b"(\\53)", 0), (b"\x2B", 5))
        self.assertEqual(PdfParser.get_value(b"(\\53a)", 0), (b"\x2Ba", 6))
        self.assertEqual(PdfParser.get_value(b"(\\1111)", 0), (b"\x491", 7))
        self.assertEqual(PdfParser.get_value(b" 123 (", 0), (123, 4))
        self.assertAlmostEqual(PdfParser.get_value(b" 123.4 %", 0)[0], 123.4)
        self.assertEqual(PdfParser.get_value(b" 123.4 %", 0)[1], 6)
        self.assertRaises(PdfFormatError, PdfParser.get_value, b"]", 0)
        d = PdfParser.get_value(b"<</Name (value) /N /V>>", 0)[0]
        self.assertIsInstance(d, PdfDict)
        self.assertEqual(len(d), 2)
        self.assertEqual(d.Name, "value")
        self.assertEqual(d[b"Name"], b"value")
        self.assertEqual(d.N, PdfName("V"))
        a = PdfParser.get_value(b"[/Name (value) /N /V]", 0)[0]
        self.assertIsInstance(a, list)
        self.assertEqual(len(a), 4)
        self.assertEqual(a[0], PdfName("Name"))
        s = PdfParser.get_value(
            b"<</Name (value) /Length 5>>\nstream\nabcde\nendstream<<...", 0
        )[0]
        self.assertIsInstance(s, PdfStream)
        self.assertEqual(s.dictionary.Name, "value")
        self.assertEqual(s.decode(), b"abcde")
        for name in ["CreationDate", "ModDate"]:
            for date, value in {
                b"20180729214124": "20180729214124",
                b"D:20180729214124": "20180729214124",
                b"D:2018072921": "20180729210000",
                b"D:20180729214124Z": "20180729214124",
                b"D:20180729214124+08'00'": "20180729134124",
                b"D:20180729214124-05'00'": "20180730024124",
            }.items():
                d = PdfParser.get_value(
                    b"<</" + name.encode() + b" (" + date + b")>>", 0
                )[0]
                self.assertEqual(time.strftime("%Y%m%d%H%M%S", getattr(d, name)), value)

    def test_pdf_repr(self):
        self.assertEqual(bytes(IndirectReference(1, 2)), b"1 2 R")
        self.assertEqual(bytes(IndirectObjectDef(*IndirectReference(1, 2))), b"1 2 obj")
        self.assertEqual(bytes(PdfName(b"Name#Hash")), b"/Name#23Hash")
        self.assertEqual(bytes(PdfName("Name#Hash")), b"/Name#23Hash")
        self.assertEqual(
            bytes(PdfDict({b"Name": IndirectReference(1, 2)})), b"<<\n/Name 1 2 R\n>>"
        )
        self.assertEqual(
            bytes(PdfDict({"Name": IndirectReference(1, 2)})), b"<<\n/Name 1 2 R\n>>"
        )
        self.assertEqual(pdf_repr(IndirectReference(1, 2)), b"1 2 R")
        self.assertEqual(
            pdf_repr(IndirectObjectDef(*IndirectReference(1, 2))), b"1 2 obj"
        )
        self.assertEqual(pdf_repr(PdfName(b"Name#Hash")), b"/Name#23Hash")
        self.assertEqual(pdf_repr(PdfName("Name#Hash")), b"/Name#23Hash")
        self.assertEqual(
            pdf_repr(PdfDict({b"Name": IndirectReference(1, 2)})),
            b"<<\n/Name 1 2 R\n>>",
        )
        self.assertEqual(
            pdf_repr(PdfDict({"Name": IndirectReference(1, 2)})), b"<<\n/Name 1 2 R\n>>"
        )
        self.assertEqual(pdf_repr(123), b"123")
        self.assertEqual(pdf_repr(True), b"true")
        self.assertEqual(pdf_repr(False), b"false")
        self.assertEqual(pdf_repr(None), b"null")
        self.assertEqual(pdf_repr(b"a)/b\\(c"), br"(a\)/b\\\(c)")
        self.assertEqual(
            pdf_repr([123, True, {"a": PdfName(b"b")}]), b"[ 123 true <<\n/a /b\n>> ]"
        )
        self.assertEqual(pdf_repr(PdfBinary(b"\x90\x1F\xA0")), b"<901FA0>")



from .helper import hopper, PillowTestCase

from PIL import Image, XVThumbImagePlugin

TEST_FILE = "Tests/images/hopper.p7"


class TestFileXVThumb(PillowTestCase):
    def test_open(self):
        # Act
        im = Image.open(TEST_FILE)

        # Assert
        self.assertEqual(im.format, "XVThumb")

        # Create a Hopper image with a similar XV palette
        im_hopper = hopper().quantize(palette=im)
        self.assert_image_similar(im, im_hopper, 9)

    def test_unexpected_eof(self):
        # Test unexpected EOF reading XV thumbnail file
        # Arrange
        bad_file = "Tests/images/hopper_bad.p7"

        # Act / Assert
        self.assertRaises(SyntaxError, XVThumbImagePlugin.XVThumbImageFile, bad_file)

    def test_invalid_file(self):
        # Arrange
        invalid_file = "Tests/images/flower.jpg"

        # Act / Assert
        self.assertRaises(
            SyntaxError, XVThumbImagePlugin.XVThumbImageFile, invalid_file
        )



#
# The Python Imaging Library.
#
# SPIDER image file handling
#
# History:
# 2004-08-02    Created BB
# 2006-03-02    added save method
# 2006-03-13    added support for stack images
#
# Copyright (c) 2004 by Health Research Inc. (HRI) RENSSELAER, NY 12144.
# Copyright (c) 2004 by William Baxter.
# Copyright (c) 2004 by Secret Labs AB.
# Copyright (c) 2004 by Fredrik Lundh.
#

##
# Image plugin for the Spider image format.  This format is is used
# by the SPIDER software, in processing image data from electron
# microscopy and tomography.
##

#
# SpiderImagePlugin.py
#
# The Spider image format is used by SPIDER software, in processing
# image data from electron microscopy and tomography.
#
# Spider home page:
# https://spider.wadsworth.org/spider_doc/spider/docs/spider.html
#
# Details about the Spider image format:
# https://spider.wadsworth.org/spider_doc/spider/docs/image_doc.html
#

from __future__ import print_function

from PIL import Image, ImageFile
import os
import struct
import sys


def isInt(f):
    try:
        i = int(f)
        if f - i == 0:
            return 1
        else:
            return 0
    except (ValueError, OverflowError):
        return 0


iforms = [1, 3, -11, -12, -21, -22]


# There is no magic number to identify Spider files, so just check a
# series of header locations to see if they have reasonable values.
# Returns no. of bytes in the header, if it is a valid Spider header,
# otherwise returns 0


def isSpiderHeader(t):
    h = (99,) + t  # add 1 value so can use spider header index start=1
    # header values 1,2,5,12,13,22,23 should be integers
    for i in [1, 2, 5, 12, 13, 22, 23]:
        if not isInt(h[i]):
            return 0
    # check iform
    iform = int(h[5])
    if iform not in iforms:
        return 0
    # check other header values
    labrec = int(h[13])  # no. records in file header
    labbyt = int(h[22])  # total no. of bytes in header
    lenbyt = int(h[23])  # record length in bytes
    if labbyt != (labrec * lenbyt):
        return 0
    # looks like a valid header
    return labbyt


def isSpiderImage(filename):
    with open(filename, "rb") as fp:
        f = fp.read(92)  # read 23 * 4 bytes
    t = struct.unpack(">23f", f)  # try big-endian first
    hdrlen = isSpiderHeader(t)
    if hdrlen == 0:
        t = struct.unpack("<23f", f)  # little-endian
        hdrlen = isSpiderHeader(t)
    return hdrlen


class SpiderImageFile(ImageFile.ImageFile):

    format = "SPIDER"
    format_description = "Spider 2D image"
    _close_exclusive_fp_after_loading = False

    def _open(self):
        # check header
        n = 27 * 4  # read 27 float values
        f = self.fp.read(n)

        try:
            self.bigendian = 1
            t = struct.unpack(">27f", f)  # try big-endian first
            hdrlen = isSpiderHeader(t)
            if hdrlen == 0:
                self.bigendian = 0
                t = struct.unpack("<27f", f)  # little-endian
                hdrlen = isSpiderHeader(t)
            if hdrlen == 0:
                raise SyntaxError("not a valid Spider file")
        except struct.error:
            raise SyntaxError("not a valid Spider file")

        h = (99,) + t  # add 1 value : spider header index starts at 1
        iform = int(h[5])
        if iform != 1:
            raise SyntaxError("not a Spider 2D image")

        self._size = int(h[12]), int(h[2])  # size in pixels (width, height)
        self.istack = int(h[24])
        self.imgnumber = int(h[27])

        if self.istack == 0 and self.imgnumber == 0:
            # stk=0, img=0: a regular 2D image
            offset = hdrlen
            self._nimages = 1
        elif self.istack > 0 and self.imgnumber == 0:
            # stk>0, img=0: Opening the stack for the first time
            self.imgbytes = int(h[12]) * int(h[2]) * 4
            self.hdrlen = hdrlen
            self._nimages = int(h[26])
            # Point to the first image in the stack
            offset = hdrlen * 2
            self.imgnumber = 1
        elif self.istack == 0 and self.imgnumber > 0:
            # stk=0, img>0: an image within the stack
            offset = hdrlen + self.stkoffset
            self.istack = 2  # So Image knows it's still a stack
        else:
            raise SyntaxError("inconsistent stack header values")

        if self.bigendian:
            self.rawmode = "F;32BF"
        else:
            self.rawmode = "F;32F"
        self.mode = "F"

        self.tile = [("raw", (0, 0) + self.size, offset, (self.rawmode, 0, 1))]
        self.__fp = self.fp  # FIXME: hack

    @property
    def n_frames(self):
        return self._nimages

    @property
    def is_animated(self):
        return self._nimages > 1

    # 1st image index is zero (although SPIDER imgnumber starts at 1)
    def tell(self):
        if self.imgnumber < 1:
            return 0
        else:
            return self.imgnumber - 1

    def seek(self, frame):
        if self.istack == 0:
            raise EOFError("attempt to seek in a non-stack file")
        if not self._seek_check(frame):
            return
        self.stkoffset = self.hdrlen + frame * (self.hdrlen + self.imgbytes)
        self.fp = self.__fp
        self.fp.seek(self.stkoffset)
        self._open()

    # returns a byte image after rescaling to 0..255
    def convert2byte(self, depth=255):
        (minimum, maximum) = self.getextrema()
        m = 1
        if maximum != minimum:
            m = depth / (maximum - minimum)
        b = -m * minimum
        return self.point(lambda i, m=m, b=b: i * m + b).convert("L")

    # returns a ImageTk.PhotoImage object, after rescaling to 0..255
    def tkPhotoImage(self):
        from PIL import ImageTk

        return ImageTk.PhotoImage(self.convert2byte(), palette=256)

    def _close__fp(self):
        try:
            if self.__fp != self.fp:
                self.__fp.close()
        except AttributeError:
            pass
        finally:
            self.__fp = None


# --------------------------------------------------------------------
# Image series

# given a list of filenames, return a list of images
def loadImageSeries(filelist=None):
    """create a list of :py:class:`~PIL.Image.Image` objects for use in a montage"""
    if filelist is None or len(filelist) < 1:
        return

    imglist = []
    for img in filelist:
        if not os.path.exists(img):
            print("unable to find %s" % img)
            continue
        try:
            im = Image.open(img).convert2byte()
        except Exception:
            if not isSpiderImage(img):
                print(img + " is not a Spider image file")
            continue
        im.info["filename"] = img
        imglist.append(im)
    return imglist


# --------------------------------------------------------------------
# For saving images in Spider format


def makeSpiderHeader(im):
    nsam, nrow = im.size
    lenbyt = nsam * 4  # There are labrec records in the header
    labrec = 1024 / lenbyt
    if 1024 % lenbyt != 0:
        labrec += 1
    labbyt = labrec * lenbyt
    hdr = []
    nvalues = int(labbyt / 4)
    for i in range(nvalues):
        hdr.append(0.0)

    if len(hdr) < 23:
        return []

    # NB these are Fortran indices
    hdr[1] = 1.0  # nslice (=1 for an image)
    hdr[2] = float(nrow)  # number of rows per slice
    hdr[5] = 1.0  # iform for 2D image
    hdr[12] = float(nsam)  # number of pixels per line
    hdr[13] = float(labrec)  # number of records in file header
    hdr[22] = float(labbyt)  # total number of bytes in header
    hdr[23] = float(lenbyt)  # record length in bytes

    # adjust for Fortran indexing
    hdr = hdr[1:]
    hdr.append(0.0)
    # pack binary data into a string
    hdrstr = []
    for v in hdr:
        hdrstr.append(struct.pack("f", v))
    return hdrstr


def _save(im, fp, filename):
    if im.mode[0] != "F":
        im = im.convert("F")

    hdr = makeSpiderHeader(im)
    if len(hdr) < 256:
        raise IOError("Error creating Spider header")

    # write the SPIDER header
    fp.writelines(hdr)

    rawmode = "F;32NF"  # 32-bit native floating point
    ImageFile._save(im, fp, [("raw", (0, 0) + im.size, 0, (rawmode, 0, 1))])


def _save_spider(im, fp, filename):
    # get the filename extension and register it with Image
    ext = os.path.splitext(filename)[1]
    Image.register_extension(SpiderImageFile.format, ext)
    _save(im, fp, filename)


# --------------------------------------------------------------------


Image.register_open(SpiderImageFile.format, SpiderImageFile)
Image.register_save(SpiderImageFile.format, _save_spider)

if __name__ == "__main__":

    if len(sys.argv) < 2:
        print("Syntax: python SpiderImagePlugin.py [infile] [outfile]")
        sys.exit()

    filename = sys.argv[1]
    if not isSpiderImage(filename):
        print("input image must be in Spider format")
        sys.exit()

    im = Image.open(filename)
    print("image: " + str(im))
    print("format: " + str(im.format))
    print("size: " + str(im.size))
    print("mode: " + str(im.mode))
    print("max, min: ", end=" ")
    print(im.getextrema())

    if len(sys.argv) > 2:
        outfile = sys.argv[2]

        # perform some image operation
        im = im.transpose(Image.FLIP_LEFT_RIGHT)
        print(
            "saving a flipped version of %s as %s "
            % (os.path.basename(filename), outfile)
        )
        im.save(outfile, SpiderImageFile.format)



from .helper import PillowTestCase, hopper
from .test_imageqt import PillowQPixmapTestCase

from PIL import ImageQt

if ImageQt.qt_is_installed:
    from PIL.ImageQt import QPixmap


class TestToQPixmap(PillowQPixmapTestCase, PillowTestCase):
    def test_sanity(self):
        for mode in ("1", "RGB", "RGBA", "L", "P"):
            data = ImageQt.toqpixmap(hopper(mode))

            self.assertIsInstance(data, QPixmap)
            self.assertFalse(data.isNull())

            # Test saving the file
            tempfile = self.tempfile("temp_{}.png".format(mode))
            data.save(tempfile)



from .helper import PillowTestCase

from PIL import Image


class TestImagePutAlpha(PillowTestCase):
    def test_interface(self):

        im = Image.new("RGBA", (1, 1), (1, 2, 3, 0))
        self.assertEqual(im.getpixel((0, 0)), (1, 2, 3, 0))

        im = Image.new("RGBA", (1, 1), (1, 2, 3))
        self.assertEqual(im.getpixel((0, 0)), (1, 2, 3, 255))

        im.putalpha(Image.new("L", im.size, 4))
        self.assertEqual(im.getpixel((0, 0)), (1, 2, 3, 4))

        im.putalpha(5)
        self.assertEqual(im.getpixel((0, 0)), (1, 2, 3, 5))

    def test_promote(self):

        im = Image.new("L", (1, 1), 1)
        self.assertEqual(im.getpixel((0, 0)), 1)

        im.putalpha(2)
        self.assertEqual(im.mode, "LA")
        self.assertEqual(im.getpixel((0, 0)), (1, 2))

        im = Image.new("P", (1, 1), 1)
        self.assertEqual(im.getpixel((0, 0)), 1)

        im.putalpha(2)
        self.assertEqual(im.mode, "PA")
        self.assertEqual(im.getpixel((0, 0)), (1, 2))

        im = Image.new("RGB", (1, 1), (1, 2, 3))
        self.assertEqual(im.getpixel((0, 0)), (1, 2, 3))

        im.putalpha(4)
        self.assertEqual(im.mode, "RGBA")
        self.assertEqual(im.getpixel((0, 0)), (1, 2, 3, 4))

    def test_readonly(self):

        im = Image.new("RGB", (1, 1), (1, 2, 3))
        im.readonly = 1

        im.putalpha(4)
        self.assertFalse(im.readonly)
        self.assertEqual(im.mode, "RGBA")
        self.assertEqual(im.getpixel((0, 0)), (1, 2, 3, 4))



from .helper import unittest, PillowTestCase

try:
    from PIL import FpxImagePlugin
except ImportError:
    olefile_installed = False
else:
    olefile_installed = True


@unittest.skipUnless(olefile_installed, "olefile package not installed")
class TestFileFpx(PillowTestCase):
    def test_invalid_file(self):
        # Test an invalid OLE file
        invalid_file = "Tests/images/flower.jpg"
        self.assertRaises(SyntaxError, FpxImagePlugin.FpxImageFile, invalid_file)

        # Test a valid OLE file, but not an FPX file
        ole_file = "Tests/images/test-ole-file.doc"
        self.assertRaises(SyntaxError, FpxImagePlugin.FpxImageFile, ole_file)
